"""
WSGI config for demo project.

It exposes the WSGI callable as a module-level variable named ``application``.

For more information on this file, see
https://docs.djangoproject.com/en/2.2/howto/deployment/wsgi/
"""

import os
import sys

from django.core.wsgi import get_wsgi_application

# This allows easy placement of apps within the interior demo directory.
app_path = os.path.abspath(  # noqa: PTH100
    os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir)  # noqa: PTH100, PTH118, PTH120
)
sys.path.append(os.path.join(app_path, "demo"))  # noqa: PTH118

# We defer to a DJANGO_SETTINGS_MODULE already in the environment. This breaks
# if running multiple sites in the same mod_wsgi process. To fix this, use
# mod_wsgi daemon mode with each site in its own daemon process, or use
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "demo.settings")

# This application object is used by any WSGI server configured to use this file.
application = get_wsgi_application()

# Apply WSGI middleware here.



from __future__ import annotations

from copy import copy
from typing import Any

from .base import Block


class Button(Block):
    """
    Render a ``<button>`` with Bootstrap styling.

    Example:

        .. code-block:: python

            from wildewidgets import Button

            button = Button(text="My Button")

        When rendered in the template with the ``wildewdigets`` template tag, this
        will produce:

        .. code-block:: html

            <button type="button" class="button btn btn-secondary">My Button</button>

    Keyword Args:
        text: The text to use for the button
        color: The Boostrap color class to use for the button
        close: The Boostrap close icon will be used for the button
        size: The Bootstrap button size - None, 'sm', 'lg'
        disabled: If ``True``, the button will be disabled

    """

    block: str = "button"
    tag: str = "button"

    #: The Boostrap color class to use for the button
    color: str = "secondary"
    #: The text to use for the button
    text: str = "Button"
    #: If ``True``, ignore :py:attr:`text` and make this into a Bootstrap
    #: "close" button with a close icon.
    close: bool = False
    #: If ``True``, ignore :py:attr:`text` and make this into a Bootstrap
    #: "close" button with a close icon.
    size: str | None = None
    disabled: bool = False

    def __init__(
        self,
        text: str | None = None,
        color: str | None = None,
        close: bool | None = None,
        size: str | None = None,
        disabled: bool | None = None,
        **kwargs: Any,
    ) -> None:
        self.color = color if color is not None else self.color
        self.text = text if text is not None else self.text
        self.close = close if close is not None else self.close
        self.size = size if size is not None else self.size
        self.disabled = disabled if disabled is not None else self.disabled
        if self.disabled:
            if "attributes" in kwargs:
                kwargs["attributes"]["disabled"] = True
            else:
                kwargs["attributes"] = {"disabled": True}
        if self.close:
            self.text = ""
        super().__init__(self.text, **kwargs)
        self._attributes["type"] = "button"
        if self.close:
            self.add_class("btn-close")
            self._aria_attributes["label"] = "Close"
        else:
            self.add_class("btn")
            self.add_class(f"btn-{self.color}")
            if self.size:
                self.add_class(f"btn-{self.size}")


class ModalButton(Button):
    """
    Render a ``<button>`` with Bootstrap styling which toggles a Bootstrap modal.

    Example:

        .. code-block:: python

            from wildewidgets import ModalButton

            button = ModalButton(text="My Button", target='#mymodal')

        When rendered in the template with the ``wildewdigets`` template tag, this
        will produce:

        .. code-block:: html

            <button type="button" class="button button--modal btn btn-secondary"
            data-toggle="modal" data-target="#mymodal">My Button</button>

    Keyword Args:
        target: The CSS target for the Bootstrap modal

    """

    block: str = "button button--modal"

    #: The CSS target for the Bootstrap modal
    target: str | None = None

    def __init__(self, target: str | None = None, **kwargs: Any) -> None:
        self.target = target or self.target
        assert self.target is not None, "ModalButton requires a target"  # noqa: S101
        super().__init__(**kwargs)
        self._data_attributes["toggle"] = "modal"
        self._data_attributes["target"] = self.target


class CollapseButton(Button):
    """
    Render a ``<button>`` with Bootstrap styling which toggles a ``collapse``.

    Example:
        ... code-block:: python

            from wildewidgets import CollapseButton

            button = CollapseButton(text="My Button", target='#mymodal')

        When rendered in the template with the ``wildewdigets`` template tag, this
        will produce:

        ... code-block:: html

            <button type="button" class="button button--collapse btn btn-secondary"
                data-toggle="collapse" data-target="#mymodal"
                aria-expanded="false" aria-controls="mymodal">My Button</button>

    Keyword Args:
        target: The CSS target for the Bootstrap collapse

    """

    block: str = "button button--collapse"

    #: The CSS target for the Bootstrap collapse
    target: str | None = None

    def __init__(self, target: str | None = None, **kwargs: Any) -> None:
        self.target = target or self.target
        assert self.target is not None, "CollapseButton requires a target"  # noqa: S101
        super().__init__(**kwargs)
        self._data_attributes["toggle"] = "collapse"
        self._data_attributes["target"] = self.target
        self._aria_attributes["expanded"] = "false"
        self._aria_attributes["controls"] = self.target.lstrip("#")


class LinkButton(Button):
    """
    Render an ``<a>`` with Bootstrap button styling which toggles a Bootstrap
    modal.

    Example:
        ... code-block:: python

            from wildewidgets import LinkButton

            button = LinkButton(text="My Button", url='https://myexample.com')

        When rendered in the template with the ``wildewdigets`` template tag, this
        will produce:

        ... code-block:: html

            <a href="https://myexample.com"
                class="button button--link btn btn-secondary">
                My Button
            </a>


    Keyword Args:
        url: The URL for the ``href`` attribute of the <a> record, defaults to no URL.

    """

    block: str = "button button--link"
    tag: str = "a"

    url: str | None = None

    def __init__(self, url: str | None = None, **kwargs: Any) -> None:
        self.url = url or self.url
        assert self.url is not None, "LinkButton requires a url"  # noqa: S101
        super().__init__(**kwargs)
        del self._attributes["type"]
        if url:
            self._attributes["href"] = self.url


class InputButton(Button):
    """
    Render an ``<input type="submit">`` with Bootstrap button styling.

    Example::

        .. code-block:: python

            from wildewidgets import InputButton

            button = InputButton(text="Save")

        When rendered in the template with the ``wildewdigets`` template tag, this
        will produce:

        .. code-block:: html

            <input type="submit"
                class="button button--submit btn btn-secondary" value="Save">

    """

    template_name: str = "wildewidgets/block--simple.html"

    block: str = "button--submit"
    tag: str = "input"

    confirm_text: str | None = None

    def __init__(self, confirm_text: str | None = None, **kwargs: Any) -> None:
        self.confirm_text = confirm_text or self.confirm_text
        super().__init__(**kwargs)
        self._attributes["type"] = "submit"
        self._attributes["value"] = self.text
        if self.confirm_text:
            self._attributes["onclick"] = f"return confirm('{self.confirm_text}');"
        self.text = ""


class FormButton(Block):
    """
    Render a ``<form>`` with optional hidden inputs and a submit button.

    Example::

        .. code-block:: python

            from wildewidgets import FormButton

            button = FormButton(
                text="Save",
                action='/my/form/action',
                data={'field1': 'value1'}
            )

        When rendered in the template with the ``wildewdigets`` template tag, this
        will produce:

        .. code-block:: html

            <form action="/my/form/action" method="post" class="button-form">
                <input type="hidden" name="csrfmiddlewaretoken"
                    value="__THE_CSRF_TOKEN__">
                <input type="hidden" name="field1" value="value1">
                <input type="submit" class="button button--submit btn btn-secondary"
                    value="Save">
            </form>

    All the constructor parameters can be set in a subclass of this class as
    class attributes.  Parameters to the constructor override any defined class
    attributes.

    Keyword Args:
        text: The text to use for the button, defaults to 'Button'
        color: The Boostrap color class to use for the button, defaults to 'secondary'
        button_css_class: a string of classes to apply to the button, defaults
            to no classes.
        button_attributes: Set any additional attributes for the button as key,
            value pairs, defaults to no additional attributes.
        button_data_attributes: Set ``data-`` attributes for the button, defaults to no
            data attributes

    """

    template_name: str = "wildewidgets/button--form.html"
    tag: str = "form"
    block: str = "button-form"

    action: str | None = None
    method: str = "post"
    text: str | None = None
    color: str = "secondary"
    button_css_class: str | None = None
    button_attributes: dict[str, str] = {}  # noqa: RUF012
    button_data_attributes: dict[str, str] = {}  # noqa: RUF012
    confirm_text: str | None = None
    data: dict[str, Any] = {}  # noqa: RUF012

    def __init__(self, **kwargs: Any) -> None:
        action = kwargs.pop("action", self.action)
        method = kwargs.pop("method", self.method)
        self.close = kwargs.pop("close", False)
        self.data = kwargs.pop("data", copy(self.data))
        self.button = self.__build_button(kwargs)
        super().__init__(**kwargs)
        self._attributes["method"] = method
        self._attributes["action"] = action

    def __build_button(self, kwargs: dict[str, Any]) -> InputButton:
        """
        Build the submit button for the form.

        Creates an InputButton with appropriate styling and attributes based on
        the configuration provided to the FormButton.

        Args:
            kwargs: Dictionary of keyword arguments to process

        Returns:
            InputButton: The configured submit button for the form

        """
        button_kwargs = {
            "text": kwargs.pop("text", self.text),
            "color": kwargs.pop("color", self.color),
            "css_class": kwargs.pop("button_css_class", self.button_css_class),
            "attributes": kwargs.pop("button_attributes", copy(self.button_attributes)),
            "data_attributes": kwargs.pop(
                "button_data_attributes", copy(self.button_data_attributes)
            ),
        }
        if self.close:
            button_kwargs["close"] = True
            button_kwargs["aria_attributes"] = {"label": "Close"}
        button_kwargs["confirm_text"] = kwargs.pop("confirm_text", self.confirm_text)
        return InputButton(**button_kwargs)

    def get_context_data(self, *args, **kwargs) -> dict[str, Any]:
        context = super().get_context_data(*args, **kwargs)
        context["button"] = self.button
        context["data"] = self.data
        return context


class ButtonRow(Block):
    """
    Renders a row of buttons with right alignment.

    This class creates a horizontal row of buttons with right-aligned (end)
    justification using Bootstrap's flexbox utilities. It's useful for placing
    action buttons at the bottom of forms or dialogs.

    Note:
        This class is planned for deprecation in favor of
        :py:class:`HorizontalLayoutBlock`.

    Example:
        .. code-block:: python

            from wildewidgets import Button, ButtonRow

            save_button = Button(text="Save", color="primary")
            cancel_button = Button(text="Cancel")
            button_row = ButtonRow(save_button, cancel_button)

        When rendered, this produces:

        .. code-block:: html

            <div class="d-flex justify-content-end">
                <button type="button" class="button btn btn-primary">Save</button>
                <button type="button" class="button btn btn-secondary">Cancel</button>
            </div>

    Args:
        *blocks: Button instances or other blocks to include in the row

    Keyword Args:
        **kwargs: Additional keyword arguments passed to the parent Block class

    """

    def __init__(self, *blocks: Any, **kwargs: Any) -> None:
        """
        Initialize a ButtonRow with buttons and styling.

        Adds flexbox utility classes to create a right-aligned row of buttons.

        Args:
            *blocks: Button instances or other blocks to include in the row
            **kwargs: Additional keyword arguments including css_class

        """
        css_class = kwargs.get("css_class", "")
        css_class += " d-flex justify-content-end"
        kwargs["css_class"] = css_class
        super().__init__(*blocks, **kwargs)



from __future__ import annotations

import warnings
from typing import Any

from .base import Block, TemplateWidget
from .buttons import CollapseButton, FormButton, LinkButton, ModalButton


class BasicHeader(TemplateWidget):
    """
    The base header class for all header widgets in the wildewidgets library.

    This class provides the core functionality for rendering HTML headers with
    various styling options, including optional badges. All other header classes
    inherit from this class.

    Example:
        .. code-block:: python

            from wildewidgets import BasicHeader

            header = BasicHeader(
                header_text="Welcome to the Dashboard",
                header_level=1,
                badge_text="New",
                badge_class="success"
            )

    Attributes:
        template_name: Path to the Django template used for rendering
        header_level: HTML heading level (1-6) to use
        header_type: Type of header styling ("h" for standard, "display" for larger)
        header_text: The text content of the header
        css_class: CSS classes to apply to the header container
        css_id: Optional ID attribute for the header element
        badge_text: Optional text for a badge beside the header
        badge_class: Bootstrap color class for the badge (e.g., "primary",
            "warning")
        badge_rounded_pill: Whether to style the badge as a rounded pill

    Keyword Args:
        header_level: the header depth, as in h1, h2, h3...
        header_type: the Boostrap 5 header class, either h-*, or display-*.
        header_text: the text of the header.
        css_class: any css classes to apply to the header.
        css_id: an ID to add to the header div.
        badge_text: text to add in a badge to the right of the main header text.
        badge_class: the css class to add to the badge.

    """

    template_name: str = "wildewidgets/header_with_controls.html"

    header_level: int = 1
    header_type: str = "h"
    header_text: str | None = None
    css_class: str = "my-3"
    css_id: str | None = None
    badge_text: str | None = None
    badge_class: str = "warning"
    badge_rounded_pill: bool = True

    def __init__(
        self,
        header_level: int | None = None,
        header_type: str | None = None,
        header_text: str | None = None,
        css_class: str | None = None,
        css_id: str | None = None,
        badge_text: str | None = None,
        badge_class: str | None = None,
        badge_rounded_pill: bool | None = None,
        **kwargs: Any,
    ):
        self.header_level = header_level or self.header_level
        self.header_type = header_type or self.header_type
        self.header_text = header_text or self.header_text
        self.css_class = css_class or self.css_class
        self.css_id = css_id or self.css_id
        self.badge_text = badge_text or self.badge_text
        self.badge_class = badge_class or self.badge_class
        self.badge_rounded_pill = (
            badge_rounded_pill
            if badge_rounded_pill is not None
            else self.badge_rounded_pill
        )
        kwargs["title"] = self.header_text
        super().__init__(**kwargs)

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        """
        Prepare the context data for template rendering.

        Processes header settings and adds them to the template context.
        Sets the appropriate header class based on the header type and level,
        and includes all other header properties.

        Args:
            **kwargs: Initial context dictionary

        Returns:
            dict: Updated context dictionary with header properties

        """
        if self.header_type == "h":
            kwargs["header_class"] = f"h{self.header_level}"
        elif self.header_type == "display":
            kwargs["header_class"] = f"display-{self.header_level}"

        kwargs["header_level"] = self.header_level
        kwargs["header_text"] = self.header_text
        kwargs["header_type"] = self.header_type
        kwargs["css_class"] = self.css_class
        kwargs["css_id"] = self.css_id
        kwargs["badge_text"] = self.badge_text
        kwargs["badge_class"] = self.badge_class
        kwargs["badge_rounded_pill"] = self.badge_rounded_pill
        return kwargs


class HeaderWithLinkButton(BasicHeader):
    """
    A header with an attached link button.

    This combines a heading element with a button that links to a specified URL.
    The button appears to the right of the header text.

    Deprecated:
        This class is deprecated since version 0.13.45.
        Use :py:class:`HeaderWithWidget` with the :py:meth:`add_link_button`
        method instead.

    Example:
        .. code-block:: python

            from wildewidgets import HeaderWithLinkButton

            header = HeaderWithLinkButton(
                header_text="Users",
                header_level=2,
                url="/users/add/",
                link_text="Add User",
                button_class="success"
            )

    Attributes:
        template_name: Path to the template for rendering this widget
        url: The URL that the button links to
        link_text: The text displayed on the button
        button_class: Bootstrap contextual class for the button

    """

    template_name = "wildewidgets/header_with_link_button.html"

    url: str | None = None
    link_text: str | None = None
    button_class: str = "primary"

    def __init__(
        self,
        url: str | None = None,
        link_text: str | None = None,
        button_class: str | None = None,
        **kwargs: Any,
    ):
        warnings.warn(
            "HeaderWithLinkButton is deprecated. Use "
            "wildewidgets.HeaderWithWidget instead",
            DeprecationWarning,
            stacklevel=2,
        )
        self.url = url or self.url
        self.link_text = link_text or self.link_text
        self.button_class = button_class or self.button_class
        super().__init__(**kwargs)

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        """
        Prepare the context data for template rendering.

        Adds link button specific variables to the context dictionary.

        Args:
            **kwargs: Initial context dictionary from parent class

        Returns:
            dict: Updated context with url, link_text, and button_class

        """
        kwargs = super().get_context_data(**kwargs)
        kwargs["url"] = self.url
        kwargs["link_text"] = self.link_text
        kwargs["button_class"] = self.button_class
        return kwargs


class HeaderWithFormButton(BasicHeader):
    """
    A header with an attached form submission button.

    This combines a heading element with a button that submits a form
    when clicked. The form must be defined elsewhere in your template.

    Deprecated:
        This class is deprecated since version 0.13.45.
        Use :py:class:`HeaderWithWidget` with the :py:meth:`add_form_button`
        method instead.

    Example:
        .. code-block:: python

            from wildewidgets import HeaderWithFormButton

            header = HeaderWithFormButton(
                header_text="Dangerous Action",
                header_level=3,
                url="/api/delete-all/",
                button_text="Delete All"
            )

    Attributes:
        template_name: Path to the template for rendering this widget
        url: The URL that the form will submit to
        button_text: The text displayed on the button

    """

    template_name = "wildewidgets/header_with_form_button.html"

    url: str | None = None
    button_text: str | None = None

    def __init__(
        self, url: str | None = None, button_text: str | None = None, **kwargs: Any
    ):
        warnings.warn(
            "HeaderWithFormButton is deprecated. Use "
            "wildewidgets.HeaderWithWidget instead",
            DeprecationWarning,
            stacklevel=2,
        )
        self.url = url or self.url
        self.button_text = button_text or self.button_text
        super().__init__(**kwargs)

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        """
        Prepare the context data for template rendering.

        Adds form button specific variables to the context dictionary.

        Args:
            **kwargs: Initial context dictionary from parent class

        Returns:
            dict: Updated context with url and button_text

        """
        kwargs = super().get_context_data(**kwargs)
        kwargs["url"] = self.url
        kwargs["button_text"] = self.button_text
        return kwargs


class HeaderWithCollapseButton(BasicHeader):
    """
    A header with a button that toggles a Bootstrap collapsible element.

    This combines a heading element with a button that expands or collapses
    a target element when clicked. The collapsible element must be defined
    elsewhere in your template.

    Deprecated:
        This class is deprecated since version 0.13.45.
        Use :py:class:`HeaderWithWidget` with the :py:meth:`add_collapse_button`
        method instead.

    Example:
        .. code-block:: python

            from wildewidgets import HeaderWithCollapseButton

            header = HeaderWithCollapseButton(
                header_text="Advanced Options",
                header_level=3,
                collapse_id="advanced-options",
                button_text="Toggle"
            )

    Attributes:
        template_name: Path to the template for rendering this widget
        collapse_id: ID of the collapsible element to toggle
        button_text: Text displayed on the button
        button_class: Bootstrap contextual class for the button

    """

    template_name: str = "wildewidgets/header_with_collapse_button.html"

    collapse_id: str | None = None
    button_text: str | None = None
    button_class: str | None = "primary"

    def __init__(
        self,
        collapse_id: str | None = None,
        button_text: str | None = None,
        button_class: str | None = None,
        **kwargs: Any,
    ):
        warnings.warn(
            "HeaderWithCollapseButton is deprecated. Use "
            "wildewidgets.HeaderWithWidget instead",
            DeprecationWarning,
            stacklevel=2,
        )
        self.collapse_id = collapse_id or self.collapse_id
        self.button_text = button_text or self.button_text
        self.button_class = button_class or self.button_class
        super().__init__(**kwargs)

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        """
        Prepare the context data for template rendering.

        Adds collapse button specific variables to the context dictionary.

        Args:
            **kwargs: Initial context dictionary from parent class

        Returns:
            dict: Updated context with collapse_id, button_text, and button_class

        """
        kwargs = super().get_context_data(**kwargs)
        kwargs["collapse_id"] = self.collapse_id
        kwargs["button_text"] = self.button_text
        kwargs["button_class"] = self.button_class
        return kwargs


class HeaderWithModalButton(BasicHeader):
    """
    A header widget with a button that triggers a Bootstrap modal.

    This class combines a heading element with a button that opens a modal dialog
    when clicked. The modal must be defined elsewhere in your template.

    Deprecated:
        This class is deprecated since version 0.13.45.
        Use :py:class:`HeaderWithWidget` with the :py:meth:`add_modal_button`

    Attributes:
        template_name: Path to the template for rendering this widget
        modal_id: ID of the modal to be triggered by the button
        button_text: Text to display on the button
        button_class: Bootstrap button class (e.g., "primary", "secondary")

    Keyword Args:
        **kwargs: Keyword arguments including modal_id, button_text, and
            button_class, plus any arguments for BasicHeader

    """

    template_name = "wildewidgets/header_with_modal_button.html"
    modal_id: str | None = None
    button_text: str | None = None
    button_class: str = "primary"

    def __init__(self, **kwargs: Any):
        warnings.warn(
            "Use wildewidgets.HeaderWithWidget instead",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(**kwargs)
        self.modal_id = kwargs.get("modal_id", self.modal_id)
        self.button_text = kwargs.get("button_text", self.button_text)
        self.button_class = kwargs.get("button_class", self.button_class)

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        """
        Prepare the context data for template rendering.

        Adds modal button specific variables to the context dictionary.

        Args:
            **kwargs: Initial context dictionary from parent class

        Returns:
            dict: Updated context with modal_id, button_text, and button_class

        """
        kwargs = super().get_context_data(**kwargs)
        kwargs["modal_id"] = self.modal_id
        kwargs["button_text"] = self.button_text
        kwargs["button_class"] = self.button_class
        return kwargs


class HeaderWithWidget(BasicHeader):
    """
    A header widget that can contain another widget (typically a button).

    This flexible header class allows adding any widget (most commonly buttons)
    next to the header text. This is the recommended approach for creating headers
    with interactive elements.

    Example:
        .. code-block:: python

            from wildewidgets import HeaderWithWidget, LinkButton

            # Create a header with a link button
            header = HeaderWithWidget(
                header_text="Products List",
                header_level=2,
                widget=LinkButton(text="Add Product", url="/products/add/")
            )

            # Or add the button after initialization
            header = HeaderWithWidget(header_text="Products List")
            header.add_link_button(text="Add Product", url="/products/add/")

    Args:
        widget: Widget to display next to the header text

    Keyword Args:
        header_text: Text content of the header
        header_level: HTML heading level (1-6)
        **kwargs: Additional arguments passed to BasicHeader

    """

    template_name = "wildewidgets/header_with_widget.html"

    def __init__(self, widget: Block | None = None, **kwargs: Any):
        super().__init__(**kwargs)
        self.widget = widget

    def add_form_button(self, **kwargs: Any) -> None:
        """
        Add a form submission button to the header.

        This method creates a FormButton instance and adds it to the header.
        The button will submit a form when clicked.

        Args:
            **kwargs: Arguments passed to :py:class:`wildewidgets.FormButton`
             constructor.
                Common arguments include:

                * text: Button label text
                * action: Form submission URL
                * color: Bootstrap color class (default: "secondary")
                * method: HTTP method (default: "post")
                * confirm_text: Text for confirmation dialog (if needed)

        """
        self.widget = FormButton(**kwargs)

    def add_link_button(self, **kwargs: Any) -> None:
        """
        Add a hyperlink button to the header.

        This method creates a LinkButton instance and adds it to the header.
        The button will navigate to the specified URL when clicked.

        Args:
            **kwargs: Arguments passed to :py:class:`wildewidgets.LinkButton`
                constructor.
                Common arguments include:

                * text: Button label text
                * url: Target URL for the link
                * color: Bootstrap color class (default: "secondary")
                * css_class: Additional CSS classes

        """
        self.widget = LinkButton(**kwargs)

    def add_modal_button(self, **kwargs: Any) -> None:
        """
        Add a button that triggers a Bootstrap modal dialog.

        This method creates a ModalButton instance and adds it to the header.
        The button will open the specified modal when clicked.

        Args:
            **kwargs: Arguments passed to :py:class:`wildewidgets.ModalButton`
                constructor.
                Common arguments include:

                * text: Button label text
                * target: Modal ID to target (with # prefix)
                * color: Bootstrap color class (default: "secondary")

        """
        self.widget = ModalButton(**kwargs)

    def add_collapse_button(self, **kwargs: Any) -> None:
        """
        Add a button that toggles a Bootstrap collapsible element.

        This method creates a CollapseButton instance and adds it to the header.
        The button will toggle the visibility of the specified collapsible element.

        Args:
            **kwargs: Arguments passed to
                :py:class:`wildewidgets.CollapseButton` constructor.
                Common arguments include:

                * text: Button label text
                * target: Collapse target ID (with # prefix)
                * color: Bootstrap color class (default: "secondary")

        """
        self.widget = CollapseButton(**kwargs)

    def set_widget(self, widget: Block) -> None:
        """
        Set or replace the widget displayed next to the header.

        This method allows setting any widget (not just buttons) to appear next
        to the header text. This provides flexibility beyond the specialized
        button methods.

        Args:
            widget: Any Block-derived widget to display in the header

        """
        self.widget = widget

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        """
        Get the context data for rendering the template.

        Adds the widget to the template context, along with other header properties.

        Args:
            **kwargs: Additional context variables

        Returns:
            dict: Updated context dictionary including the widget

        """
        kwargs = super().get_context_data(**kwargs)
        kwargs["widget"] = self.widget
        return kwargs


class PageHeader(HeaderWithWidget):
    """
    Provides a standard page header.

    This is an H1 header with a widget, typically used at the top of a page
    to provide a title and an action button or other interactive element.

    Example:
        .. code-block:: python

            from wildewidgets import PageHeader, LinkButton

            header = PageHeader(
                header_text="Welcome to the Dashboard",
                widget=LinkButton(text="Learn More", url="/about/")
            )

    """

    css_class: str = "my-4"


class CardHeader(HeaderWithWidget):
    """
    A header specifically designed for Bootstrap card components.

    This provides an H2 header with appropriate styling for use as a card header,
    with the option to include an action button or other widget.

    Example:
        .. code-block:: python

            from wildewidgets import CardHeader, LinkButton

            header = CardHeader(
                header_text="User Details",
                widget=LinkButton(text="Edit", url="/users/123/edit/")
            )

    """

    css_class: str = "my-3"
    header_level: int = 2


class WidgetListLayoutHeader(HeaderWithWidget):
    """
    A specialized header for use with WidgetListLayout components.

    This header is designed to match the styling of the WidgetListLayout
    component, providing consistent margins and header size. It can include
    an optional action button or other widget.

    Example:
        .. code-block:: python

            from wildewidgets import WidgetListLayoutHeader, LinkButton

            header = WidgetListLayoutHeader(
                header_text="Dashboard Widgets",
                widget=LinkButton(text="Add Widget", url="/widgets/add/")
            )

    """

    css_class: str = "mb-4"
    header_level: int = 2



from django.urls import include, path
from wildewidgets import WildewidgetDispatch

from .core import urls as core_urls

urlpatterns = [
    path("wildewidgets_json", WildewidgetDispatch.as_view(), name="wildewidgets_json"),
    path("", include(core_urls, namespace="core")),
]



from __future__ import annotations

import warnings
from typing import Any

try:
    from pygments import highlight
    from pygments.formatters import HtmlFormatter
    from pygments.lexers import get_lexer_by_name
except ModuleNotFoundError:
    # Only needed if using syntax highlighting
    pass

from .base import Block, TemplateWidget


class CodeWidget(Block):
    """
    A widget that displays code with syntax highlighting.

    This widget uses Pygments to apply syntax highlighting to code snippets.
    The programming language must be specified to determine the highlighting rules.

    Example:
        .. code-block:: python

            from wildewidgets import CodeWidget

            python_code = '''
            def hello_world():
                print("Hello, world!")
            '''

            code_widget = CodeWidget(
                code=python_code,
                language="python",
                line_numbers=True
            )

    Note:
        Requires the `pygments` package to be installed.

    Args:
        code: The code to be displayed
        language: The programming language for syntax highlighting
        line_numbers: Whether to display line numbers
        **kwargs: Additional arguments passed to the parent class

    Raises:
        ValueError: If no language is specified
        RuntimeError: If Pygments is not installed

    """

    #: The name of the block for CSS styling
    block: str = "wildewidgets_highlight_container"

    #: The code to be displayed
    code: str = ""
    #: The programming language for the :py:attr:`code`
    language: str | None = None
    #: If ``True``, show line numbers
    line_numbers: bool = False

    def __init__(
        self,
        code: str | None = None,
        language: str | None = None,
        line_numbers: bool = False,
        **kwargs: Any,
    ):
        # Raise RuntimeError if pygments is not installed
        if not (get_lexer_by_name and highlight and HtmlFormatter):
            msg = "Pygments is not installed. Please install it to use CodeWidget."
            raise RuntimeError(msg)
        self.code = code or self.code
        self.language = language or self.language
        if not self.language:
            msg = (
                f'{self.__class__.__name__}: "language" must be defined either as a '
                "class attribute or a keyword arg"
            )
            raise ValueError(msg)
        self.line_numbers = line_numbers or self.line_numbers
        super().__init__(**kwargs)
        self.add_code(self.code, language=self.language, line_numbers=self.line_numbers)

    def add_code(self, code: str, language: str, line_numbers: bool = False) -> None:
        """
        Apply syntax highlighting to code and add it to the widget.

        Args:
            code: The code to highlight
            language: The programming language to use for highlighting
            line_numbers: Whether to display line numbers

        """
        lexer = get_lexer_by_name(language)
        formatter = HtmlFormatter(
            linenos=line_numbers, cssclass="wildewidgets_highlight"
        )
        self.add_block(highlight(code, lexer, formatter))


class MarkdownWidget(TemplateWidget):
    """
    A widget that renders Markdown text as HTML.

    This widget converts Markdown formatted text to HTML and displays it.
    It requires Django's template system to render the HTML.

    Attributes:
        template_name: The Django template used to render the widget
        text: The Markdown text to render
        css_class: CSS classes to apply to the container

    Example:
        .. code-block:: python

            from wildewidgets import MarkdownWidget

            markdown_text = '''
            # Hello World

            This is a **bold** statement with *emphasis*.

            - Item 1
            - Item 2
            '''

            widget = MarkdownWidget(text=markdown_text, css_class="my-markdown")

    Args:
        *args: Positional arguments passed to the parent class

    Keyword Args:
        **kwargs: Keyword arguments including:
            text: The Markdown text to render
            css_class: CSS classes to apply to the container

    """

    template_name = "wildewidgets/markdown_widget.html"
    text: str = ""
    css_class: str = ""

    def __init__(self, *args: Any, **kwargs: Any):
        self.text = kwargs.pop("text", self.text)
        self.css_class = kwargs.pop("css_class", self.css_class)
        super().__init__(*args, **kwargs)

    def get_context_data(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """
        Prepare the context data for template rendering.

        Args:
            *args: Positional arguments passed to the parent method
            **kwargs: Keyword arguments passed to the parent method

        Returns:
            dict: The context dictionary with markdown text and CSS classes

        """
        kwargs = super().get_context_data(*args, **kwargs)
        kwargs["text"] = self.text
        kwargs["css_class"] = self.css_class
        return kwargs


class HTMLWidget(TemplateWidget):
    """
    A widget that renders raw HTML content.

    This widget allows for displaying arbitrary HTML content within
    your Django application. Use with caution as it can introduce
    security risks if displaying user-provided content.

    Attributes:
        template_name: The Django template used to render the widget
        html: The raw HTML content to display
        css_class: CSS classes to apply to the container

    Example:
        .. code-block:: python

            from wildewidgets import HTMLWidget

            html_content = '<p>This is <strong>HTML</strong> content</p>'
            widget = HTMLWidget(html=html_content, css_class="custom-html-block")

    Warning:
        Be careful when using this widget with user-provided content as it
        could lead to cross-site scripting (XSS) vulnerabilities.

    Args:
        *args: Positional arguments passed to the parent class

    Keyword Args:
        **kwargs: Keyword arguments including:
            html: The raw HTML content to display
            css_class: CSS classes to apply to the container

    """

    template_name = "wildewidgets/html_widget.html"
    html: str = ""
    css_class: str | None = None

    def __init__(self, *args: Any, **kwargs: Any):
        self.html = kwargs.pop("html", self.html)
        self.css_class = kwargs.pop("css_class", self.css_class)
        super().__init__(*args, **kwargs)

    def get_context_data(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """
        Prepare the context data for template rendering.

        Args:
            *args: Positional arguments passed to the parent method
            **kwargs: Keyword arguments passed to the parent method

        Returns:
            dict: The context dictionary with HTML content and CSS classes

        """
        kwargs = super().get_context_data(*args, **kwargs)
        kwargs["html"] = self.html
        kwargs["css_class"] = self.css_class
        return kwargs


class StringBlock(Block):
    """
    A basic widget that displays a string.

    This is a simple wrapper around Block that initializes with a text string.

    .. deprecated:: 0.14.0
        Use :py:class:`wildewidgets.Block` directly instead.
        It works exactly like :py:class:`StringBlock`

    Args:
        text: The text to display
        **kwargs: Additional arguments passed to the parent class

    Example:
        .. code-block:: python

            # Deprecated usage
            from wildewidgets import StringBlock

            text_block = StringBlock("Hello, world!")

            # Preferred usage
            from wildewidgets import Block

            text_block = Block("Hello, world!")

    Args:
        text: The text to display

    Keyword Args:
        **kwargs: Additional arguments passed to the parent class

    """

    def __init__(self, text: str, **kwargs: Any):
        warnings.warn(
            "Deprecated in 0.14.0; use Block directly instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(*[text], **kwargs)


class TimeStamp(Block):
    """
    A widget that displays text formatted as a timestamp.

    This widget renders text in a small font with light styling,
    making it suitable for timestamps, captions, or other secondary text.

    Attributes:
        tag: HTML tag to use for the timestamp (small)
        css_class: CSS class for styling the timestamp

    Example:
        .. code-block:: python

            from wildewidgets import TimeStamp

            timestamp = TimeStamp("Last updated: 2023-06-15 14:30")

    """

    tag = "small"
    css_class = "fw-light"


class TagBlock(Block):
    """
    A widget that displays text as a colored badge/tag.

    This widget creates a Bootstrap badge with customizable color,
    useful for status indicators, categories, or other labeled items.

    Example:
        .. code-block:: python

            from wildewidgets import TagBlock

            # Creates a success (green) badge
            status_tag = TagBlock("Active", color="success")

            # Creates a danger (red) badge
            error_tag = TagBlock("Error", color="danger")

            # Creates a custom colored badge
            custom_tag = TagBlock("Custom", color="info")

    Note:
        The color parameter accepts standard Bootstrap color names like
        "primary", "secondary", "success", "danger", "warning", "info", etc.

    Args:
        text: The text to display in the badge

    Keyword Args:
        color: The Bootstrap color class to use (default: "secondary")
        **kwargs: Additional arguments passed to the parent class


    """

    block: str = "badge"

    def __init__(self, text: str, color: str = "secondary", **kwargs: Any):
        super().__init__(text, **kwargs)
        self.add_class(f"bg-{color}")



from __future__ import annotations

import random
from copy import deepcopy
from typing import Any

from django import template

from wildewidgets.views import WidgetInitKwargsMixin

from ..base import Widget
from .components import (
    DataTableColumn,
    DataTableFilter,
    DataTableForm,
    DataTableStyler,
)
from .views import DatatableAJAXView


class BaseDataTable(Widget, WidgetInitKwargsMixin, DatatableAJAXView):  # type: ignore[misc]
    """
    Base class for creating interactive `dataTables <https://datatables.net/>`_
    with sorting, filtering, and pagination.

    This class provides the foundation for building powerful data tables with
    features like:

    - Client-side or server-side processing
    - Column sorting and filtering
    - Pagination
    - Custom styling and formatting
    - Action buttons for row operations
    - Bulk actions through form submissions

    The `BaseDataTable` can operate in two modes:

    1. Synchronous mode: All data is loaded and processed in the browser
    2. Asynchronous mode: Data is loaded via AJAX from the server as needed

    Example:
        .. code-block:: python

            from wildewidgets import BaseDataTable

            class UserTable(BaseDataTable):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    self.add_column("username", "Username")
                    self.add_column("email", "Email")
                    self.add_column("date_joined", "Joined")

                def render_date_joined_column(self, row, column):
                    return row.date_joined.strftime("%Y-%m-%d")

    Args:
        *args: Positional arguments passed to the parent Widget class

    Keyword Args:
        title: The title of the table, displayed above the table controls
        width: CSS width for the table (default is "100%")
        height: CSS height for the table (optional)
        searchable: Whether to show a search input (default is True)
        paging: Whether to enable pagination (default is True)
        page_length: Number of rows per page (default is 25)
        small: Use smaller font and row height (default is False)
        buttons: Add export buttons (default is False)
        striped: Use alternating row colors (default is False)
        hide_controls: Hide pagination and search controls (default is False)
        table_id: Custom CSS ID for the table (random if None)
        sort_ascending: Default sort order for the table (default is True)
        data: Initial data to populate the table (optional, empty by default)
        form_actions: List of bulk actions for selected rows (optional)
        form_url: URL to submit form actions (optional)
        ajax_url_name: URL name for the AJAX endpoint (default is "wildewidgets_json")
        column_wrap_fields: List of fields to wrap in table controls (optional)
        is_async: If True, use AJAX to load data when the table is empty
        (default is True)
        has_select_all: If True and form_actions is set, show a "select all"
            checkbox in the checkbox column header (default is False)
        **kwargs: Keyword arguments for parent Widget class initialization


    """

    #: The Django template file to use for rendering the table
    template_file: str = "wildewidgets/table.html"

    #: The URL name for the dataTables AJAX endpoint
    ajax_url_name: str = "wildewidgets_json"

    #: The header fields to wrap in the table controls
    column_wrap_fields: list[str] = []  # noqa: RUF012

    # dataTable specific configs

    #: How tall should we make the table?  Any CSS width string is valid.
    height: str | None = None
    #: How wide should we make the table?  Any CSS width string is valid.
    width: str | None = "100%"
    #: If ``True``, use smaller font and row height when rendering rows
    small: bool = False
    #: If ``True``, use different colors for even and odd rows
    striped: bool = False
    #: Show the search input?
    searchable: bool = True
    #: How many rows should we show on each page
    page_length: int = 25
    #: If ``True``, sort rows ascending; otherwise descending.
    sort_ascending: bool = True
    #: Hide our paging, page length and search controls
    hide_controls: bool = False

    #: The CSS id to assign to the table.  The id will be ``datatable_table_{table_id}``
    #: If this is ``None``, a random id will be generated.
    table_id: str | None = None

    #: If ``True``, add the dataTable "Copy", "CSV", "Export" and "Print" buttons
    buttons: bool = False

    #: Whole table form actions.  If this is not ``None``, add a first column with
    #: checkboxes to each row, and a form that allows you to choose bulk actions
    #: to perform on all checked rows.
    form_actions = None
    #: If ``True`` and :py:attr:`form_actions` is set, show a "select all" checkbox
    #: in the checkbox column header. Default is ``False`` so existing tables are
    #: unchanged.
    has_select_all: bool = False
    #: The URL to which to POST our form actions if :py:attr:`form_actions` is
    #: not ``None``
    form_url: str = ""
    #: If ``True``, use AJAX to load data when the table is empty
    is_async: bool = True

    def __init__(  # noqa: PLR0913
        self,
        *args,
        width: str | None = None,
        height: str | None = None,
        title: str | None = None,
        searchable: bool | None = True,
        paging: bool | None = True,
        page_length: int | None = None,
        small: bool | None = None,
        buttons: bool | None = None,
        striped: bool | None = None,
        hide_controls: bool | None = None,
        table_id: str | None = None,
        sort_ascending: bool | None = None,
        data: list[Any] | None = None,
        form_actions: Any = None,
        form_url: str | None = None,
        ajax_url_name: str | None = None,
        column_wrap_fields: list[str] | None = None,
        is_async: bool | None = None,
        has_select_all: bool | None = None,
        **kwargs,
    ):
        self.width = width if width is not None else self.width
        self.height = height if height is not None else self.height
        self.title = title if title is not None else self.title
        self.searchable = searchable if searchable is not None else self.searchable
        self.page_length = page_length if page_length is not None else self.page_length
        self.small = small if small is not None else self.small
        self.buttons = buttons if buttons is not None else self.buttons
        self.striped = striped if striped is not None else self.striped
        self.hide_controls = (
            hide_controls if hide_controls is not None else self.hide_controls
        )
        self.has_select_all = (
            has_select_all if has_select_all is not None else self.has_select_all
        )
        #: These are options for dataTable itself and get set in the JavaScript
        #: constructor for the table.
        self.datatable_options: dict[str, Any] = {
            "width": self.width,
            "height": self.height,
            "title": self.title,
            "searchable": self.searchable,
            "paging": paging,
            "page_length": self.page_length,
            "small": self.small,
            "buttons": self.buttons,
            "striped": self.striped,
            "hide_controls": self.hide_controls,
            "has_select_all": self.has_select_all,
        }
        #: The CSS id for this table
        self.table_id = table_id or self.table_id
        if self.table_id is None:
            self.table_id = str(random.randrange(0, 1000))  # noqa: S311
        self.table_name = f"datatable_table_{self.table_id}"
        # We have to do this this way instead of naming it above in the kwargs
        # because ``async`` is a reserved keyword
        self.async_if_empty: bool = is_async if is_async is not None else self.is_async
        #: A mapping of field name to column definition
        self.column_fields: dict[str, DataTableColumn] = {}
        #: A mapping of field name to column filter definition
        self.column_filters: dict[str, DataTableFilter] = {}
        #: A list of column styles to apply
        self.column_styles: list[DataTableStyler] = []
        self.data = data or []
        self.sort_ascending = (
            sort_ascending if sort_ascending is not None else self.sort_ascending
        )
        self.form_actions = form_actions or deepcopy(self.form_actions)
        self.form_url = form_url or self.form_url
        self.ajax_url_name = ajax_url_name or self.ajax_url_name
        self.column_wrap_fields = column_wrap_fields or deepcopy(
            self.column_wrap_fields
        )
        if self.has_form_actions():
            self.column_fields["checkbox"] = DataTableColumn(
                field="checkbox",
                verbose_name=" ",
                searchable=False,
                sortable=False,
            )
        super().__init__(*args, **kwargs)

    def get_column_number(self, name: str) -> int:
        """
        Get the numerical index of a column in the table by its name.

        This is useful when you need to reference columns in JavaScript operations
        or when configuring DataTables-specific functionality.

        Args:
            name: The field name of the column to find

        Returns:
            int: Zero-based index of the column in the table

        Raises:
            IndexError: If no column with the given name exists in the table

        """
        columns = list(self.column_fields.keys())
        if name in columns:
            return columns.index(name)
        msg = f'No column with name "{name}" is registered with this table'
        raise IndexError(msg)

    def has_form_actions(self) -> bool:
        """
        Check if this table has form actions defined.

        Form actions allow users to perform bulk operations on selected rows
        by checking the checkboxes and submitting the form.

        Returns:
            bool: True if form actions are defined, False otherwise

        """
        return self.form_actions is not None

    def get_form_actions(self) -> Any:
        """
        Get the list of form actions defined for this table.

        Returns:
            Any: The :py:attr:`form_actions` list or ``None`` if no actions are
                defined

        """
        return self.form_actions

    def add_form_action(self, action: Any) -> None:
        """
        Add a new form action to this table.

        Form actions appear in a dropdown when rows are selected via checkboxes.

        Args:
            action: The form action to add (usually a tuple of label and handler)

        """
        if not self.form_actions:
            self.form_actions = []
        self.form_actions.append(action)

    def add_column(
        self,
        field: str,
        verbose_name: str | None = None,
        searchable: bool = True,
        sortable: bool = True,
        align: str = "left",
        head_align: str = "left",
        visible: bool = True,
        wrap: bool = True,
    ) -> None:
        """
        Add a column to the table definition.

        This method defines a new column in the table. The table will look for a
        method named ``render_{field}_column`` to handle custom rendering of the
        column's cell values.

        Args:
            field: The name of the field/attribute to render in this column
            verbose_name: Display name for the column header (defaults to
                capitalized field name)
            searchable: Whether this column is included in global search
            sortable: Whether the table can be sorted by this column
            align: Horizontal alignment for cell content ("left", "right", "center")
            head_align: Horizontal alignment for the header cell ("left",
                "right", "center")
            visible: Whether the column is initially visible
            wrap: Whether to wrap content in this column

        Example:
            .. code-block:: python

                from wildewidgets import BaseDataTable

                table = BaseDataTable(
                    title="User List",
                    searchable=True,
                    paging=True,
                    page_length=10
                )

                table.add_column(
                    "created_at",
                    "Created",
                    align="right",
                    sortable=True
                )

        """
        self.column_fields[field] = DataTableColumn(
            field=field,
            verbose_name=verbose_name,
            searchable=searchable,
            sortable=sortable,
            align=align,
            head_align=head_align,
            visible=visible,
            wrap=wrap,
        )

    def add_filter(self, field: str, dt_filter: DataTableFilter) -> None:
        """
        Add a filter control for a specific column.

        Filters allow users to narrow down data based on column values.

        Args:
            field: The name of the field/column to filter
            dt_filter: A DataTableFilter instance defining the filter behavior

        Example:
            .. code-block:: python

                from wildewidgets import BaseDataTable, DataTableFilter

                table = BaseDataTable(
                    title="User List",
                    searchable=True,
                    paging=True,
                    page_length=10
                )

                table.add_filter(
                    "status",
                    DataTableFilter(
                        choices=[("active", "Active"), ("inactive", "Inactive")]
                    )
                )

        """
        self.column_filters[field] = dt_filter

    def remove_filter(self, field: str) -> None:
        """
        Remove a previously defined filter from a column.

        Args:
            field: The name of the field/column to remove the filter from

        """
        del self.column_filters[field]

    def add_styler(self, styler: DataTableStyler) -> None:
        """
        Add a style rule to the table.

        Stylers allow conditional formatting of cells based on their values
        or the values of other cells in the same row.

        Args:
            styler: A :py:class:`wildewidgets.DataTableStyler` instance defining
                the styling rule

        Example:
            .. code-block:: python

                from wildewidgets import BaseDataTable, DataTableStyler

                table = BaseDataTable(
                    title="User List",
                    searchable=True,
                    paging=True,
                    page_length=10
                )

                # Style the "status" column red when its value is "error"
                table.add_styler(
                    DataTableStyler(
                        test_cell="status",
                        test_value="error",
                        css_class="text-danger"
                    )
                )

        """
        styler.test_index = list(self.column_fields.keys()).index(styler.test_cell)
        if styler.target_cell:
            styler.target_index = list(self.column_fields.keys()).index(
                styler.target_cell
            )
        self.column_styles.append(styler)

    def build_context(self, **kwargs) -> dict[str, Any]:
        """
        Build the context for synchronous table rendering.

        This method adds the table's row data to the context for template
        rendering. Override this method to customize how data is prepared
        for the template.

        Args:
            **kwargs: The template context to update

        Returns:
            dict: The updated context with row data

        """
        kwargs["rows"] = self.data
        return kwargs

    def get_template_context_data(self, **kwargs) -> dict[str, Any]:
        """
        Prepare the complete context for table template rendering.

        This method builds the context dictionary with all necessary data for
        rendering the table template, including configuration, filters,
        headers, and data mode (async or sync).

        Args:
            **kwargs: Initial context values

        Returns:
            dict: Complete context dictionary for template rendering

        """
        kwargs = super().get_template_context_data(**kwargs)
        has_filters = False
        has_filter_defaults = False
        filters: list[tuple[DataTableColumn, DataTableFilter] | None] = []
        for key, item in self.column_fields.items():
            if key in self.column_filters:
                column_filter = self.column_filters[key]
                filters.append((item, column_filter))
                has_filters = True
                if column_filter.default:
                    has_filter_defaults = True
            else:
                filters.append(None)
        if self.data or not self.async_if_empty:
            kwargs = self.build_context(**kwargs)
            kwargs["async"] = False
        else:
            kwargs["async"] = True
        kwargs["header"] = self.column_fields
        kwargs["has_form_actions"] = self.has_form_actions()
        kwargs["filters"] = filters
        kwargs["stylers"] = self.column_styles
        kwargs["has_filters"] = has_filters
        kwargs["has_filter_defaults"] = has_filter_defaults
        kwargs["options"] = self.datatable_options
        table_id = self.table_id or str(random.randrange(0, 1000))  # noqa: S311
        kwargs["name"] = f"datatable_table_{table_id}"
        kwargs["sort_ascending"] = self.sort_ascending
        kwargs["column_wrap_fields"] = self.column_wrap_fields
        kwargs["tableclass"] = self.__class__.__name__
        if not self.data:
            kwargs["extra_data"] = self.get_encoded_extra_data()
        kwargs["form"] = DataTableForm(self)
        if "csrf_token" in kwargs:
            kwargs["csrf_token"] = kwargs["csrf_token"]
        kwargs["ajax_url_name"] = self.ajax_url_name
        return kwargs

    def get_content(self, **kwargs) -> str:
        """
        Render the table to HTML.

        This method generates the complete HTML for the table by rendering
        the template with the prepared context.

        Args:
            **kwargs: Additional context values to include

        Returns:
            str: The rendered HTML for the table

        """
        context = self.get_template_context_data(**kwargs)
        html_template = template.loader.get_template(self.template_file)  # type: ignore[attr-defined]
        return html_template.render(context)

    def __str__(self) -> str:
        """
        Return the string representation of the table.

        This method allows the table to be included directly in templates
        using the {{ table }} syntax.

        Returns:
            str: The rendered HTML for the table

        """
        return self.get_content()

    def add_row(self, **kwargs) -> None:
        """
        Add a row to the table's data.

        This method is used for synchronous tables to add data directly.

        Args:
            **kwargs: Field values as key-value pairs, where keys match column names

        Example:
            .. code-block:: python

                from wildewidgets import BaseDataTable

                table = BaseDataTable(
                    title="User List",
                    searchable=True,
                    paging=True,
                    page_length=10
                )

                table.add_row(
                    username="johndoe",
                    email="john@example.com",
                    date_joined="2023-01-15"
                )

        """
        row = []
        for field in self.column_fields:
            if field in kwargs:
                row.append(kwargs[field])
            else:
                row.append("")
        self.data.append(row)

    def render_checkbox_column(self, row: Any, column: str) -> str:  # noqa: ARG002
        """
        Render the checkbox column for form actions.

        This method generates the HTML for the checkbox in each row when
        form_actions are enabled. The checkbox uses the row's ID as its value.

        Args:
            row: The data object for the current row
            column: The column name (always "checkbox")

        Returns:
            str: HTML for the checkbox input

        """
        return f'<input type="checkbox" name="checkbox" value="{row.id}">'



__version__ = "1.4.4"

from .forms import *  # noqa: F403
from .menus import *  # noqa: F403
from .views import *  # noqa: F403
from .widgets import *  # noqa: F403
# WARNING: do not import .viewsets here, because it required Django to be
# fully initialized.  That makes it impossible to build the docs.



# Configuration file for the Sphinx documentation builder.  # noqa: INP001
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#

import os
import sys

import django

sys.path.insert(0, os.path.abspath("../demo"))
os.environ.setdefault("TESTING", "True")
os.environ["DJANGO_SETTINGS_MODULE"] = "demo.settings"
django.setup()


# -- Project information -----------------------------------------------------

# the master toctree document
master_doc = "index"

project = "django-wildewidgets"
copyright = "2023, California Institute of Technology"  # noqa: A001
author = "Glenn Bach, Chris Malek"

# The full version, including alpha/beta/rc tags
release = "1.4.4"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinxcontrib.images",
    "sphinx.ext.intersphinx",
]


# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

add_function_parentheses: bool = False
add_module_names: bool = False

autodoc_member_order = "groupwise"

# Make Sphinx not expand all our Type Aliases
autodoc_type_aliases: dict[str, str] = {}

# the locations and names of other projects that should be linked to this one
intersphinx_mapping: dict[str, tuple[str, str | None]] = {
    "python": ("https://docs.python.org/3", None),
    "django": (
        "http://docs.djangoproject.com/en/dev/",
        "http://docs.djangoproject.com/en/dev/_objects/",
    ),
}

# Configure the path to the Django settings module
django_settings = "demo.settings_docker"
# Include the database table names of Django models
django_show_db_tables = True

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"
html_context = {
    "display_github": True,  # Integrate github
    "github_user": "caltech-imss-ads",  # Username
    "github_repo": "django-wildewidgets",  # Repo name
    "github_version": "main",  # Version
    "conf_py_path": "/docs/",  # Path in the checkout to the docs root
}

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    "collapse_navigation": True,
    "navigation_depth": 3,
    "show_prev_next": False,
    "logo": {
        "image_light": "wildewidgets_logo.png",
        "image_dark": "wildewidgets_dark_mode_logo.png",
        "text": "Django-Wildewidgets",
    },
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/caltechads/django-wildewidgets",
            "icon": "fab fa-github-square",
            "type": "fontawesome",
        },
        {
            "name": "Demo",
            "url": "https://wildewidgets.caltech.edu",
            "icon": "fa fa-desktop",
            "type": "fontawesome",
        },
    ],
}
# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_logo = "_static/wildewidgets.png"
html_favicon = "_static/favicon.ico"



from __future__ import annotations

import logging
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Literal, cast

from django.conf import settings
from django.core.exceptions import FieldDoesNotExist, ImproperlyConfigured
from django.db import models
from django.db.models.fields.related import ManyToManyRel, RelatedField

from .actions import (
    ActionButtonBlockMixin,
    ActionsButtonsBySpecMixin,
    RowActionButton,
)
from .base import BaseDataTable

if TYPE_CHECKING:
    import datetime

    from django.contrib.contenttypes.fields import GenericForeignKey

    from ..base import Widget

logger = logging.getLogger(__name__)


# -------------------------------
# Mixins
# -------------------------------


class ModelTableMixin:
    """
    Mixin used to create a table from a Django Model with automatic column
    configuration.

    This mixin automatically discovers model fields and creates appropriate
    table columns with sensible defaults for alignment, sorting, and display
    formatting. It handles both direct model fields and related fields through
    Django's double-underscore notation.

    Important:
        Typically you will not use this directly in your code, but instead use
        one of the pre-defined table classes like :py:class:`BasicModelTable`,
        :py:class:`ActionButtonModelTable`, or
        :py:class:`StandardActionButtonModelTable`.

    Notes:
        - Django ``DateField`` and ``DateTimeField`` fields are automatically
          formatted using the ``WILDEWIDGETS_DATE_FORMAT`` and ``
          WILDEWIDGETS_DATETIME_FORMAT`` settings.
        - Django ``BooleanField`` fields can be rendered with custom icons using
          the ``bool_icons`` attribute, which maps field names to tuples of icon
          specifications.
        - For the ``field_types`` attribute/kwarg, currently available choices for
          the dict values are:

          * ``date``: For formatting date fields
          * ``datetime``: For formatting datetime fields
          * ``currency``: For formatting currency values with a dollar sign
          * ``bool``: For rendering boolean values as check icons

        - You may add a new field type by implementing a method named
          ``render_FIELD_type_column`` on your table class, where
          ``FIELD`` is the name of the field type (e.g.,
          ``render_currency_type_column``)

    Example:
        .. code-block:: python

            from django.db import models
            from django.contrib.contenttypes.fields import GenericForeignKey
            from wildewidgets.widgets.tables import ModelTableMixin, DataTable

            class Author(models.Model):
                full_name = models.CharField(max_length=100)

            class Book(models.Model):
                title = models.CharField(max_length=200)
                authors = models.ManyToManyField(Author, related_name='books')
                isbn = models.CharField(max_length=20)
                published_date = models.DateField()

            # Example table using ModelTableMixin

            class BookTable(ModelTableMixin, DataTable):
                model = Book
                fields = ['title', 'authors__full_name', 'isbn', 'published_date']
                alignment = {'isbn': 'center', 'published_date': 'right'}
                verbose_names = {'authors__full_name': 'Authors'}
                field_types = {'published_date': 'date'}

                def render_authors__full_name_column(self, row, column):
                    return ", ".join([a.full_name for a in row.authors.all()])

    Args:
        *args: Variable length argument list (unused)

    Keyword Args:
        model: The Django model class to use for this table.  If not provided,
            the :py:attr:`model` attribute must be set on the class.
        fields: A list of field names to include in the table.  If not provided,
            defaults to all fields on the model.  The ``fields`` attribute can be set to
            ``None``, an empty list, or the string `"__all__"` to include all
            fields on the model. If a list is provided, it can include related fields
            using Django's double-underscore notation (e.g., `authors__full_name`).
        hidden: A list of field names to hide by default.  Users can unhide fields via
            the table controls.
        verbose_names: A dictionary mapping field names to custom column headers where
            the key is the field name and the value is the desired header text.
        unsortable: A list of field names that will not be sortable.  This means
            that the up/down arrows will not be displayed in the table header
            for these fields, and they will not be included in the sort query.
        unsearchable: A list of field names that will not be searched when doing
            a global table search.  This means the field will not be included
            in the search query, but it will still be displayed in the table.
        field_types: A dictionary mapping field names to data types for automatic
            formatting of table values.
        alignment: A dictionary mapping field names to alignment values. Valid
            values are ``left``, ``right``, and ``center``.  The key is the field name
            and the value is the alignment value.  If not provided, defaults to
            ``left`` for all text fields and ``right`` for numeric fields.
        bool_icons: A dictionary mapping field names to tuples of icon specifications
            for rendering boolean values. Each tuple should contain 1-2 icon specs:
            (icon_name, css_class) for True and optionally (icon_name,
            css_class) for False.

    Raises:
        ImproperlyConfigured: If the model is not set and no model is provided
            during initialization.
        django.core.exceptions.FieldDoesNotExist: If a specified field does not exist
            on the model or related models.

    """

    #: The Django model class for this table
    model: type[models.Model] | None = None

    #: This is either ``None``, the string ``__all__`` or a list of column names
    #: to use in our table.  For the list, entries can either be field names
    #: from our :py:attr:`model`, or names of computed fields that will be
    #: rendered with a ``render_FIELD_column`` method.  If ``None``, empty list
    #: or ``__all__``, display all fields on the :py:attr:`model`.
    fields: str | list[str] | None = []  # noqa: RUF012
    #: The list of field names to hide by default
    hidden: list[str] = []  # noqa: RUF012
    #: A mapping of field name to table column heading
    verbose_names: dict[str, str] = {}  # noqa: RUF012
    #: A list of field names that will not be sortable
    unsortable: list[str] = []  # noqa: RUF012
    #: A list of field names that will not be searched when doing a global table search
    unsearchable: list[str] = []  # noqa: RUF012
    #: A mapping of field name to data type.  This is used to do some automatic
    #: formatting of table values.
    field_types: dict[str, str] = {}  # noqa: RUF012
    #: A mapping of field name to field alignment.  Valid values are ``left``,
    #: ``right``, and : ``center``
    alignment: dict[str, Literal["left", "right", "center"]] = {}  # noqa: RUF012
    bool_icons: dict[str, tuple[tuple[str, str], ...]] = {}  # noqa: RUF012

    def __init__(
        self,
        *args,
        model: type[models.Model] | None = None,
        fields: str | list[str] | None = None,
        hidden: list[str] | None = None,
        verbose_names: dict[str, str] | None = None,
        unsortable: list[str] | None = None,
        unsearchable: list[str] | None = None,
        field_types: dict[str, str] | None = None,
        alignment: dict[str, Literal["left", "right", "center"]] | None = None,
        bool_icons: dict[str, tuple[tuple[str, str], ...]] | None = None,
        **kwargs,
    ):
        self.model = model if model is not None else self.model
        if not self.model:
            msg = f"{self.__class__.__name__} requires a model to be set"
            raise ImproperlyConfigured(msg)
        self.fields = fields or deepcopy(self.fields)
        self.hidden = hidden or deepcopy(self.hidden)
        self.verbose_names = verbose_names or deepcopy(self.verbose_names)
        self.unsortable = unsortable or deepcopy(self.unsortable)
        self.unsearchable = unsearchable or deepcopy(self.unsearchable)
        self.field_types = field_types or deepcopy(self.field_types)
        self.alignment = alignment or deepcopy(self.alignment)
        self.bool_icons = bool_icons or deepcopy(self.bool_icons)
        super().__init__(*args, **kwargs)

        #: A mapping of field name to Django field class
        self.model_fields: dict[
            str, models.Field | models.ForeignObjectRel | GenericForeignKey
        ] = {}
        #: A mapping of field name to Django related field class
        self.related_fields: dict[
            str, models.Field | models.ForeignObjectRel | GenericForeignKey
        ] = {}
        #: A list of names of our model fields
        self.field_names: list[str] = []

        # Build our mapping of all known fields on :py:attr:`model`
        for field in cast("models.Model", self.model)._meta.get_fields():
            if field.name == "id":
                continue
            self.model_fields[field.name] = field
            self.field_names.append(field.name)

        # Find our related fields -- these are in Django QuerySet format, e.g.
        # parent__child or parent__child__grandchild
        if isinstance(self.fields, list):
            for field_name in self.fields:
                if field_name not in self.model_fields:
                    _field = self.get_related_field(
                        cast("type[models.Model]", self.model), field_name
                    )
                    if _field:
                        self.related_fields[field_name] = _field

        if not self.fields or self.fields == "__all__":
            self.load_all_fields()
        else:
            for field_name in self.fields:
                self.load_field(field_name)

    def get_related_model(
        self, current_model: type[models.Model], field_name: str
    ) -> type[models.Model] | None:
        """
        Get the related model for a relationship field.

        Args:
            current_model: The model containing the relationship field
            field_name: The name of the relationship field

        Returns:
            The related model class or None if not found

        """
        return current_model._meta.get_field(field_name).related_model

    def get_related_field(
        self, current_model: type[models.Model], name: str
    ) -> models.Field | models.ForeignObjectRel | GenericForeignKey | None:
        """
        Resolve a field from a related model using Django's double-underscore notation.

        This method handles field paths like 'author__publisher__name' by traversing
        the model relationships to find the final field object.

        Args:
            current_model: The starting model class
            name: Field path using Django's double-underscore notation

        Returns:
            The Django Field instance for the final field in the path, or None
            if not found

        Example:
            For a path 'author__publisher__name', this would:

            1. Get the 'author' field from the current model
            2. Get the related model (Publisher)
            3. Return the 'name' field from the Publisher model

        """
        name_fields: list[str] = name.split("__")
        if len(name_fields) == 1:
            # This is not a related field specifier
            return None
        # Walk through the fields and find the Django model for the last
        # field in the spec
        for field in name_fields[:-1]:
            related_model_class = self.get_related_model(current_model, field)
            if not related_model_class:
                return None
            current_model = related_model_class
        if current_model:
            try:
                final_field = current_model._meta.get_field(name_fields[-1])
            except FieldDoesNotExist:
                final_field = None
        else:
            final_field = None
        return final_field

    def get_field(
        self, field_name: str
    ) -> models.Field | models.ForeignObjectRel | GenericForeignKey | None:
        """
        Get a field instance by name, checking both direct and related fields.

        Args:
            field_name: The name of the field to retrieve

        Returns:
            The Django Field instance or None if not found

        """
        if field_name in self.model_fields:
            field = self.model_fields[field_name]
        elif field_name in self.related_fields:
            field = self.related_fields[field_name]
        else:
            field = None
        return field

    def set_standard_column_attributes(
        self, field_name: str, kwargs: dict[str, Any]
    ) -> None:
        """
        Set standard column attributes based on field type and configuration.

        This method configures:

        - Visibility based on the 'hidden' list
        - Searchability based on the 'unsearchable' list
        - Sortability based on the 'unsortable' list
        - Alignment based on the 'alignment' dict or field type

        Args:
            field_name: The name of the field to configure
            kwargs: Dictionary of attributes to update

        """
        if field_name in self.hidden:
            kwargs["visible"] = False
        if field_name in self.unsearchable:
            kwargs["searchable"] = False
        if field_name in self.unsortable:
            kwargs["sortable"] = False
        if field_name in self.alignment:
            kwargs["align"] = self.alignment[field_name]
        else:
            field = self.get_field(field_name)
            if isinstance(
                field,
                (
                    models.TextField,
                    models.CharField,
                    models.DateField,
                    models.DateTimeField,
                ),
            ):
                kwargs["align"] = "left"
            else:
                kwargs["align"] = "right"

    def load_field(self, field_name: str) -> None:
        """
        Add a column for a single field to the table.

        This method handles field discovery, verbose name resolution, and column
        attribute configuration before adding the column to the table.

        Args:
            field_name: The name of the field to add as a column

        """
        if field_name in self.model_fields:
            field = self.model_fields[field_name]
            verbose_name = field.name.replace("_", " ")
            kwargs = {}
            if field_name in self.verbose_names:
                kwargs["verbose_name"] = self.verbose_names[field_name]
            elif hasattr(field, "verbose_name"):
                if verbose_name == field.verbose_name:
                    kwargs["verbose_name"] = verbose_name.capitalize()
                else:
                    kwargs["verbose_name"] = str(field.verbose_name)
            self.set_standard_column_attributes(field_name, kwargs)
            self.add_column(field_name, **kwargs)
        else:
            kwargs = {}
            if field_name in self.verbose_names:
                verbose_name = self.verbose_names[field_name]
            else:
                verbose_name = (
                    field_name.replace("_", " ").replace("__", " ").capitalize()
                )
            kwargs["verbose_name"] = verbose_name
            self.set_standard_column_attributes(field_name, kwargs)
            self.add_column(field_name, **kwargs)

    def load_all_fields(self) -> None:
        """
        Add columns for all discovered model fields to the table.

        This is called when fields="__all__" or when fields is empty.
        """
        for field_name in self.field_names:
            self.load_field(field_name)

    def render_currency_type_column(self, value: Any) -> str:
        """
        Format a value as currency with a dollar sign.

        Args:
            value: The value to format

        Returns:
            Formatted currency string with dollar sign

        """
        return f"${value}"

    def render_bool_type_column(self, value: Any) -> str:
        """
        Format a boolean value as a check icon for True values.

        Args:
            value: The boolean value to format

        Returns:
            HTML for a checkmark icon if True, empty string if False

        """
        if value == "True":
            return '<i class="bi-check-lg text-success"><span style="display:none">True</span></i>'  # noqa: E501
        return ""

    def render_bool_icon_column(
        self, value: Any, icon_data: tuple[tuple[str, str], ...]
    ) -> str:
        """
        Format a boolean value using custom icons.

        Args:
            value: The boolean value to format
            icon_data: Tuple of icon specifications (icon_name, css_class)

        Returns:
            HTML for the specified icon based on the boolean value

        Note:
            The icon_data tuple should contain 1-2 icon specifications:
            - First icon is used for True values
            - Second icon (if provided) is used for False values

        """
        if len(icon_data) == 0:
            return value
        if value == "True":
            return f"<i class='bi-{icon_data[0][0]} {icon_data[0][1]}'><span style='display:none'>True</span></i>"  # noqa: E501
        if len(icon_data) > 1:
            return f"<i class='bi-{icon_data[1][0]} {icon_data[1][1]}'><span style='display:none'>False</span></i>"  # noqa: E501
        return ""

    def render_datetime_type_column(self, value: datetime.datetime) -> str:
        """
        Format a datetime value with the configured datetime format.

        Args:
            value: The datetime value to format

        Returns:
            Formatted datetime string

        Note:
            The format can be customized using the `WILDEWIDGETS_DATETIME_FORMAT`
            setting in Django settings.

        """
        datetime_format = "%m/%d/%Y %H:%M"
        if hasattr(settings, "WILDEWIDGETS_DATETIME_FORMAT"):
            datetime_format = settings.WILDEWIDGETS_DATETIME_FORMAT
        if value:
            return value.strftime(datetime_format)
        return ""

    def render_date_type_column(self, value: datetime.date) -> str:
        """
        Format a date value with the configured date format.

        Args:
            value: The date value to format

        Returns:
            Formatted date string, optionally with title attribute for full date format

        Note:
            The format can be customized using the WILDEWIDGETS_DATE_FORMAT
            setting in Django settings.

        """
        date_format = "%m/%d/%Y"
        full_date_format = date_format
        modified_date_format = False
        if hasattr(settings, "WILDEWIDGETS_DATE_FORMAT"):
            date_format = settings.WILDEWIDGETS_DATE_FORMAT
            modified_date_format = True
        if value:
            date_str = value.strftime(date_format)
            if modified_date_format:
                full_date_str = value.strftime(full_date_format)
                return f"<span title='{full_date_str}'>{date_str}</span>"
            return date_str
        date_str = ""
        if modified_date_format:
            return f"<span title='{date_str}'>{date_str}</span>"
        return date_str

    def render_column(self, row: Any, column: str) -> str:
        """
        Render a cell value with appropriate formatting based on field type.

        This method handles special formatting for different field types:

        - Applies custom rendering for fields in field_types dictionary
        - Automatically formats DateTimeField and DateField values
        - Applies icon rendering for boolean fields in bool_icons dictionary

        Args:
            row: The data object for the current row
            column: The name of the column to render

        Returns:
            Formatted HTML string for the cell value

        """
        value = super().render_column(row, column)  # type: ignore[misc]
        if column in self.model_fields:
            field = self.model_fields[column]
        elif column in self.related_fields:
            field = self.related_fields[column]
        else:
            field = None
        if column in self.field_types:
            field_type = self.field_types[column]
            attr_name = f"render_{field_type}_type_column"
            if hasattr(self, attr_name):
                if attr_name in ["date", "datetime"]:
                    value = getattr(row, column)
                return getattr(self, attr_name)(value)
        elif isinstance(field, models.DateTimeField):
            return self.render_datetime_type_column(getattr(row, column))
        elif isinstance(field, models.DateField):
            return self.render_date_type_column(getattr(row, column))
        elif column in self.bool_icons:
            return self.render_bool_icon_column(value, self.bool_icons[column])
        return value


class WidgetCellMixin:
    """
    Mixin that enables using custom widgets to render individual table cells.

    This mixin allows you to specify custom widget classes for different columns,
    providing complete control over how cell data is rendered. Each widget receives
    the row data and column name, allowing for complex rendering based on the full
    context of the data.

    If a widget is defined for a column in the :py:attr:`cell_widgets` dictionary,
    it will be used to render the cell.  Otherwise, the standard column rendering
    will be used.

    Important:
        Cell widgets need to be :py:class:`wildewidgets.Block` subclasses that
        accept two additional keyword arguments:

        - row: The data object for the current row
        - column: The name of the column to render

        .. code-block:: python

            from wildewidgets import Block

            class MyCellWidget(Block):

                def __init__(self, row=None, column=None, **kwargs):
                    self.row = row
                    self.column = column
                    self.value = getattr(row, column)
                    super().__init__(f"Value is {self.value}", **kwargs)

    Attributes:
        cell_widgets: Dictionary mapping column names to widget classes for rendering
            cells in those columns

    Example:
        .. code-block:: python

            from wildewidgets import WidgetCellMixin, DataTable
            from wildewidgets.widgets.base import Block

            class StatusWidget(Block):
                def __init__(self, row=None, column=None, **kwargs):
                    super().__init__(**kwargs)
                    status = getattr(row, column)
                    if status == 'active':
                        self.add_class('badge bg-success')
                    elif status == 'pending':
                        self.add_class('badge bg-warning')
                    else:
                        self.add_class('badge bg-secondary')
                    self.text = status.capitalize()

            class MyTable(WidgetCellMixin, DataTable):
                cell_widgets = {
                    'status': StatusWidget
                }

    """

    #: Specify the widget to use for each column. The key is the column name
    #: and the value is the widget class. The widget class must be a subclass
    #: of :py:class:`wildewidgets.Block`. Fields that are not specified
    #: will not have a widget, but will be displayed as text.
    cell_widgets: dict[str, type[Widget]] = {}  # noqa: RUF012

    def render_column(self, row: Any, column: str) -> str:
        """
        Render a cell using a custom widget if defined for the column.

        If a widget is defined for the column in :py:attr:`cell_widgets`, this
        method instantiates the widget with the row and column data and returns
        its rendered output.  Otherwise, it falls back to the standard column
        rendering.

        Args:
            row: The data object for the current row
            column: The name of the column to render

        Returns:
            Rendered HTML string for the cell

        """
        if column in self.cell_widgets:
            widget_class = self.cell_widgets[column]
            widget = widget_class(row=row, column=column)
            return str(widget)
        return super().render_column(row, column)  # type: ignore[misc]


# -------------------------------
# Tables
# -------------------------------


class DataTable(ActionsButtonsBySpecMixin, BaseDataTable):  # type: ignore[misc]
    """
    Standard data table implementation with action buttons specs.

    This class combines :py:class:`BaseDataTable` with
    :py:class:`wildewidgets.ActionsButtonsBySpecMixin` to create a complete
    table implementation that supports all base table features plus action
    buttons for each row.

    Example:
        .. code-block:: python

            from wildewidgets.widgets.tables import DataTable

            class MyTable(DataTable):
                actions = [
                    ('Edit', 'my_app:edit'),
                    ('Delete', 'my_app:delete', 'post', 'danger')
                ]

                def get_queryset(self):
                    return MyModel.objects.all()

    """


class BasicModelTable(ModelTableMixin, DataTable):  # type: ignore[misc]
    """
    Ready-to-use table for displaying Django model data.

    This class combines :py:class:`ModelTableMixin` and :py:class:`DataTable` to
    provide a complete solution for displaying model data with minimal
    configuration. Simply define your model and field configuration as class
    attributes.

    Example:
        .. code-block:: python

            from django.db import models
            from wildewidgets.widgets.tables import BasicModelTable

            class Author(models.Model):
                full_name = models.CharField(max_length=100)

            class Book(models.Model):
                title = models.CharField(max_length=200)
                authors = models.ManyToManyField(Author, related_name='books')
                isbn = models.CharField(max_length=20)

            # Example table using BasicModelTable

            class BookTable(BasicModelTable):
                model = Book
                fields = ['title', 'authors__full_name', 'isbn']
                alignment = {'authors': 'left'}
                verbose_names = {'authors__full_name': 'Authors'}
                buttons = True
                striped = True

                def render_authors__full_name_column(self, row, column):
                    authors = row.authors.all()
                    if authors.count() > 1:
                        return f"{authors[0].full_name} + {authors.count()-1} more"
                    return authors[0].full_name if authors else ""

    """


class ActionButtonModelTable(ActionButtonBlockMixin, ModelTableMixin, BaseDataTable):  # type: ignore[misc]
    """
    Model table with customizable row action buttons.

    This class allows you to define action buttons for each row using
    :py:class:`RowActionButton` instances and subclasses, providing flexibility
    in button appearance and behavior. Buttons can perform different actions
    based on the row data, like linking to detail pages, opening modals, or
    submitting forms.

    Example:
        .. code-block:: python

            from django.db import models
            from wildewidgets import (
                ActionButtonModelTable,
                RowModelUrlButton,
                RowFormButton
            )

            class Author(models.Model):
                full_name = models.CharField(max_length=100)

            class Book(models.Model):
                title = models.CharField(max_length=200)
                authors = models.ManyToManyField(Author, related_name='books')
                isbn = models.CharField(max_length=20)

                def get_absolute_url(self):
                    return f"/books/{self.id}/"

                def get_delete_url(self):
                    return f"/books/{self.id}/delete/"

            class BookTable(ActionButtonModelTable):
                model = Book
                fields = ['title', 'authors__full_name', 'isbn']
                verbose_names = {'authors__full_name': 'Authors'}

                actions = [
                    RowModelUrlButton(
                        text='View',
                        color='primary',
                        attribute='get_absolute_url'
                    ),
                    RowFormButton(
                        text='Delete',
                        color='danger',
                        attribute='get_delete_url',
                        form_fields=['id'],
                        confirm_text='Are you sure you want to delete this book?'
                    )
                ]

    """


class StandardActionButtonModelTable(  # type: ignore[misc]
    ActionButtonBlockMixin, ModelTableMixin, BaseDataTable
):  # type: ignore[misc]
    """
    Model table with standard "Edit" and "Delete" buttons for each row.

    This class provides a convenient implementation for the common pattern of
    displaying model data with "Edit" and "Delete" actions. It automatically creates
    these buttons and links them to the appropriate URLs from your model.

    Requirements:
        - Your model must have the :py:class:`wildewidgets.ViewSetMixin` mixin in
          its inheritance chain
        - Your model must implement ``get_absolute_url``, ``get_update_url`` and
          ``get_delete_url`` methods

    Attributes:
        model: The Django model class to display
        fields: List of fields to include in the table

    Example:
        .. code-block:: python

            from django.db import models
            from django.urls import reverse
            from wildewidgets.models import ViewSetMixin

            class Book(ViewSetMixin, models.Model):
                title = models.CharField('Title', max_length=100)
                isbn = models.CharField('ISBN', max_length=20)

                def get_absolute_url(self):
                    return reverse('books:detail', args=[self.id])

                def get_update_url(self):
                    return reverse('books:edit', args=[self.id])

                def get_delete_url(self):
                    return reverse('books:delete', args=[self.id])

            class BookTable(StandardActionButtonModelTable):
                model = Book
                fields = ['title', 'isbn']

    """


class LookupModelTable(ActionButtonBlockMixin, ModelTableMixin, BaseDataTable):  # type: ignore[misc]
    """
    Specialized table for lookup/reference data with automatic field discovery.
    This is used by the :py:class:`wildewidgets.ModelViewSet` to display
    reference data like categories, statuses, or other lookup tables.

    This table is designed for displaying reference/lookup tables and
    automatically configures itself to:

    1. Show only direct model fields (no related fields)
    2. Left-align the ID column for better readability
    3. Support row actions via :py:class:`wildewidgets.ActionButtonBlockMixin`

    This table is ideal for admin interfaces or data management views for
    reference data like categories, statuses, or other lookup tables.

    Requirements:
        - Your model must implement ``get_absolute_url``, ``get_update_url`` and
          ``get_delete_url`` methods

    Note:
        If fields are not specified, the table automatically includes all
        non-relation fields from the model.

    Example:
        .. code-block:: python

            from django.db import models
            from wildewidgets.widgets.tables import LookupModelTable

            class Category(models.Model):
                name = models.CharField(max_length=100)
                description = models.TextField(blank=True, null=True)

                def get_absolute_url(self):
                    return f"/categories/{self.id}/"

                def get_update_url(self):
                    return f"/categories/{self.id}/edit/"

                def get_delete_url(self):
                    return f"/categories/{self.id}/delete/"

            class CategoryTable(LookupModelTable):
                model = Category

    Keyword Args:
        fields: A list of field names to include in the table. If not provided,
            defaults to all non-relation fields on the model.
        model: The Django model class to use for this table. If not provided,
            the :py:attr:`model` attribute must be set on the class.

    Raises:
        ImproperlyConfigured: If the model is not set as either a class attribute
            or provided during initialization.

    """

    actions: list[RowActionButton] = []  # noqa: RUF012

    def __init__(self, *args, **kwargs):
        fields = kwargs.pop("fields", [])
        self.model = kwargs.get("model") or self.model
        if not self.model:
            msg = f"{self.__class__.__name__} requires a model to be set"
            raise ImproperlyConfigured(msg)
        if not fields and not self.fields:
            alignment = {}
            fields = []
            for field in self.model._meta.get_fields():
                if not isinstance(field, (RelatedField, ManyToManyRel)):
                    # Exclude any RelatedField
                    fields.append(field.name)
                    if field.name == "id":
                        alignment["id"] = "left"
            kwargs["fields"] = fields
            kwargs["alignment"] = alignment
        super().__init__(*args, **kwargs)



from __future__ import annotations

from copy import copy, deepcopy
from typing import TYPE_CHECKING, Any, Literal, cast

from django.core.exceptions import ImproperlyConfigured
from django.urls import reverse

from ..base import Block
from ..buttons import InputButton
from ..forms import HiddenInputBlock
from ..structure import HorizontalLayoutBlock

if TYPE_CHECKING:
    from collections.abc import Callable

    from django.contrib.auth.models import AbstractUser


# -------------------------------
# Buttons
# -------------------------------


class RowActionButton(Block):
    """
    Base class for action buttons displayed in table rows.

    This class provides the foundation for creating buttons that appear in the
    "Actions" column of tables. Each button is bound to a specific row and can
    perform actions based on that row's data. Note that all keyword arguments/class
    attributes from :py:class:`wildewidgets.Block` are also accepted.

    You can use this class directly with keyword arguments to create buttons
    that link to row-specific URLs. In order for this to work, you'll need to
    pass a callable into `url` that accepts a single argument, which will be
    the row data. This callable should return a string that is the URL for
    the button.

    You will more usually subclass this, setting class attributes to create
    specific model-based buttons, and overriding the :py:meth:`get_url` method
    to return the URL for the button based on the row data.

    Attributes:
        tag: HTML tag for the button, defaults to "a" for links
        text: The text displayed on the button
        url: The URL the button links to, a Django URL name to reverse, or a callable
            that accepts a ``row`` arg and returns a URL string
        color: Bootstrap color class for the button (e.g., "primary", "secondary")
        size: Bootstrap size class for the button (e.g., "sm", "lg")
        permission: Optional Django permission string required to see the button
        row: Reference to the row data this button is bound to (set by bind method)

    Example:
        Direct usage with keyword arguments:

        .. code-block:: python

            from django.urls import reverse
            from wildewidgets import RowActionButton

            def get_home_url(row: Any) -> str:
                return reverse("app:home", kwargs={"id": row.id})

            button = RowActionButton(
                text="Home",
                url=get_home_url,
                color="primary",
                permission="app.view_item"
            )

        Subclassing for specific actions:

        .. code-block:: python

            class RowViewButton(RowActionButton):
                text = "View"
                color = "info"
                size = "sm"
                url = "core:item-view"

                def get_url(self, row: Any) -> str:
                    return reverse(self.url, kwargs={"id": row.id})

    Keyword Args:
        text: The text to display on the button
        url: The URL the button should link to
        color: The Bootstrap color class for the button (default: "secondary")
        size: The Bootstrap size class for the button (default: None)
        permission: Optional Django permission string required to see this
            button, e.g. "app.view_item"
        **kwargs: Additional attributes passed to the parent Block class

    Raises:
        ImproperlyConfigured: If 'text' or 'url' is not provided

    """

    tag: str = "a"

    text: str | None = None
    url: str | Callable[[Any], str] | None = None
    color: str = "secondary"
    size: str | None = None
    permission: str | None = None

    def __init__(
        self,
        text: str | None = None,
        url: str | Callable[[Any], str] | None = None,
        color: str | None = None,
        size: str | None = None,
        permission: str | None = None,
        **kwargs,
    ):
        self.text = text or self.text
        self.url = url or self.url
        self.color = color or self.color
        self.size = size or self.size
        self.permission = permission or self.permission
        super().__init__(cast("str", self.text), **kwargs)

        #: The table will set this
        self.row: Any = None

    def is_visible(self, row: Any, user: AbstractUser | None) -> bool:  # type: ignore[override]  # noqa: ARG002
        """
        Determine if this button should be visible to the given user.

        Args:
            row: The row data this button is associated with
            user: The current Django user

        Returns:
            bool: True if the button should be visible, False otherwise

        Note:
            If a permission is specified, the user must have that permission
            for the button to be visible.

        """
        if self.permission is not None and user is not None:
            return bool(user.has_perm(self.permission))
        return True

    def get_url(self, row: Any) -> str:
        """
        Get the URL for this button, potentially based on row data.

        This base implementation simply returns the static URL property.
        Subclasses can override this to generate dynamic URLs based on row data.

        Args:
            row: The row data to use for URL generation

        Returns:
            str: The URL for the button

        """
        if callable(self.url):
            # If the URL is a callable, call it with the row data
            return self.url(row)
        return self.url  # type: ignore[return-value]

    def bind(
        self,
        row: Any,
        table: ActionButtonBlockMixin,  # noqa: ARG002
        size: str | None = None,
    ) -> RowActionButton:
        """
        Bind this button to a specific row and return a copy configured for that
        row.

        This method creates a copy of the button configured specifically for the
        given row, including setting appropriate URLs, colors and sizes.

        Args:
            row: The row data to bind this button to
            table: The table instance that contains this row
            size: Optional size override for the button

        Returns:
            RowActionButton: A new button instance bound to the specified row

        Note:
            The table parameter provides access to table-level properties
            like CSRF tokens needed for form submissions.

        """
        action = deepcopy(self)
        action.row = row
        action.add_class(f"btn-{self.color}")
        action._attributes["href"] = self.get_url(row)
        if not size:
            size = self.size
        if self.size:
            action.add_class(f"btn-{self.size}")
        return action


class RowLinkButton(RowActionButton):
    """
    A row action button that renders as a link with appropriate attributes.

    This subclass of :py:class:`RowActionButton` ensures the rendered button has
    the appropriate attributes for accessibility and behavior as a link.

    Either subclass this, setting class attributes to create specific
    model-based buttons, or use it directly with keyword arguments to create
    buttons that link to model-specific URLs.  Note that all keyword
    arguments/class attributes from :py:class:`wildewidgets.Block`  and
    :py:class:`RowActionButton` are also accepted.

    Note:
        If ``row`` is going to be a Django model instance, you're better off
        using :py:class:`RowModelUrlButton` or :py:class:`RowDjangoUrlButton`
        instead.

        Honestly, the only difference between this and :py:class:`RowActionButton`
        is that we set the ``role`` attribute on the `<a>` tag to "button".

    Attributes:
        Inherits all attributes from :py:class:`RowActionButton`.

    Example:
        Direct usage with keyword arguments:

        .. code-block:: python

            from django.urls import reverse

            from wildewidgets import RowLinkButton

            def get_details_url(row: Any) -> str:
                return reverse("app:item-details", kwargs={"id": row.id})

            button = RowLinkButton(
                text="Details",
                url=get_details_url,
                color="info"
            )

        Subclassing for specific actions:

        .. code-block:: python

            from django.urls import reverse_lazy
            from wildewidgets import RowLinkButton

            class RowDetailsButton(RowLinkButton):
                text = "Details"
                color = "info"

                def get_url(self, row: Any) -> str:
                    return reverse("app:item-details", kwargs={"id": row.id})


    """

    def get_context_data(self, *args, **kwargs) -> dict[str, Any]:
        """
        Prepare the context data for template rendering.

        Sets appropriate attributes for accessibility including href and role.

        Args:
            *args: Positional arguments passed to the parent method
            **kwargs: Keyword arguments passed to the parent method

        Returns:
            dict: The context dictionary with updated attributes

        """
        self._attributes["href"] = self.get_url(self.row)
        self._attributes["role"] = "button"
        return super().get_context_data(*args, **kwargs)


class RowModelUrlButton(RowActionButton):
    """
    A row action button that gets its URL from a model instance attribute or
    method.

    This button uses a specified attribute or method of the row's model instance
    to determine its URL, making it ideal for standard model actions like view
    or edit.

    Either subclass this, setting class attributes to create specific
    model-based buttons, or use it directly with keyword arguments to create
    buttons that link to model-specific URLs.  Note that all keyword
    arguments/class attributes from :py:class:`wildewidgets.Block` and
    :py:class:`RowActionButton` are also accepted.

    Attributes:
        block: CSS class for styling ("btn")
        modifier: CSS modifier class ("auto")
        attribute: The model attribute or method to use for URL generation. Your model
            must implement this method or attribute to return the URL.
        text: Default button text ("View")

    Example:
        Direct usage with keyword arguments:

        .. code-block:: python

            from wildewidgets import RowModelUrlButton

            # Uses row.get_absolute_url() to determine URL
            button = RowModelUrlButton(text="View Details")

            # Uses row.get_edit_url() to determine URL
            button = RowModelUrlButton(
                text="Edit",
                attribute="get_edit_url",
                color="primary"
            )

        Subclassing for specific actions:

        .. code-block:: python

            class RowViewButton(RowModelUrlButton):
                attribute = "get_absolute_url"
                text = "View"
                color = "info"
                size = "sm"

    Keyword Args:
        attribute: The model attribute or method to use for URL generation. Your model
            must implement this method or attribute to return the URL.
        **kwargs: Additional keyword arguments for the
            *:py:class:`RowActionButton` class

    """

    block: str = "btn"
    modifier: str = "auto"

    attribute: str = "get_absolute_url"
    text: str | None = "View"

    def __init__(self, attribute: str | None = None, **kwargs):
        self.attribute = attribute or self.attribute
        super().__init__(**kwargs)

    def get_url(self, row: Any) -> str:
        """
        Get the URL from the specified model attribute or method.

        Args:
            row: The model instance to get the URL from

        Returns:
            str: The URL for the button

        Raises:
            ValueError: If the specified attribute doesn't exist on the model

        """
        if not hasattr(row, self.attribute):
            msg = f'{row} has no "{self.attribute}" attribute or method'
            raise ValueError(msg)
        attr = getattr(row, self.attribute)
        if callable(attr):
            return attr()
        return attr


class RowDjangoUrlButton(RowActionButton):
    """
    A row action button that constructs its URL using Django's URL resolver.

    This button uses Django's URL resolution system with either positional
    or keyword arguments to construct its URL dynamically based on row data.

    Either subclass this, setting class attributes to create specific
    model-based buttons, or use it directly with keyword arguments to create
    buttons that link to model-specific URLs.  Note that all keyword
    arguments/class attributes from :py:class:`wildewidgets.Block`  and
    :py:class:`RowActionButton` are also accepted.

    Attributes:
        block: CSS class for styling ("btn")
        url_path: The Django URL pattern name to reverse, e.g., "app:item-edit"
        url_args: List of row attribute names to use as positional arguments
        url_kwargs: Dictionary mapping URL param names to row attribute names

    Example:
        Direct usage with keyword arguments:

        .. code-block:: python

            from wildewidgets import RowDjangoUrlButton

            # Using positional arguments (e.g., 'item/1/')
            button = RowDjangoUrlButton(
                text="Edit",
                url_path="app:item-edit",
                url_args=["id"]
            )

            # Using keyword arguments (e.g., 'item/edit/?pk=1')
            button = RowDjangoUrlButton(
                text="Edit",
                url_path="app:item-edit",
                url_kwargs={"pk": "id"}
            )

        Subclassing for specific actions:

        .. code-block:: python

            class RowEditButton(RowDjangoUrlButton):
                url_path = "app:item-edit"
                text = "Edit"
                color = "primary"
                url_args = ["id"]

    Note:
        You must specify either ``url_args`` or ``url_kwargs``, but not both.

    Keyword Args:
        url_path: The Django URL pattern name to reverse
        url_args: List of row attribute names to use as positional arguments
        url_kwargs: Dictionary mapping URL param names to row attribute names

    Raises:
        ImproperlyConfigured: If both ``url_args`` and ``url_kwargs`` are provided,
            or if url_path is not set
        RowActionButton.RequiredAttrOrKwarg: If url_path is not set

    """

    block: str = "btn"

    url_path: str | None = None
    url_args: list[str] = []  # noqa: RUF012
    url_kwargs: dict[str, str] = {}  # noqa: RUF012

    def __init__(
        self,
        url_path: str | None = None,
        url_args: list[str] | None = None,
        url_kwargs: dict[str, str] | None = None,
        **kwargs,
    ):
        self.url_path = url_path or self.url_path
        self.url_args = url_args or copy(self.url_args)
        self.url_kwargs = url_kwargs or copy(self.url_kwargs)
        if not self.url_path:
            msg = "url_path"
            raise self.RequiredAttrOrKwarg(msg)
        if self.url_args and self.url_kwargs:
            msg = (
                'Either "url_args" or "url_kwargs" must be provided as keyword '
                "arguments or class attributes, but not both"
            )
            raise ImproperlyConfigured(msg)
        super().__init__(**kwargs)

    def get_url(self, row: Any) -> str:
        """
        Construct a URL using Django's URL resolver with row data.

        Args:
            row: The row data to use for URL parameters

        Returns:
            str: The resolved URL for the button

        """
        if not (self.url_args and self.url_kwargs):
            url = reverse(self.url_path)
        if self.url_args:
            args = [getattr(row, arg) for arg in self.url_args]
            url = reverse(self.url_path, args=args)
        if self.url_kwargs:
            kwargs = {kwarg: getattr(self.row, attr) for kwarg, attr in self.url_kwargs}
            url = reverse(self.url_path, kwargs=kwargs)
        return url


class RowFormButton(RowModelUrlButton):
    """
    A row action button that renders as a form, submitting via POST.

    This button renders as an HTML form that submits via POST when clicked.
    It's ideal for actions that modify data, like delete or status change
    operations.  You do not need to provide a form -- it will be created
    automatically.

    Either subclass this, setting class attributes to create specific
    model-based buttons, or use it directly with keyword arguments to create
    buttons that link to model-specific URLs.  Note that all keyword
    arguments/class attributes from :py:class:`wildewidgets.Block`  and
    :py:class:`RowModelUrlButton` are also accepted.

    Attributes:
        block: CSS class for styling ("action-form")
        tag: HTML tag for the container ("form")
        form_fields: List of row attribute names to include as hidden fields
        confirm_text: Confirmation message to show before submission

    Example:
        Direct usage with keyword arguments:

        .. code-block:: python

            from wildewidgets import RowFormButton

            button = RowFormButton(
                text="Delete",
                attribute="get_delete_url",
                form_fields=["id"],
                confirm_text="Are you sure you want to delete this item?"
            )

        Subclassing for specific actions:

        .. code-block:: python

            class RowDeleteButton(RowFormButton):
                attribute = "get_delete_url"
                text = "Delete"
                color = "outline-secondary"
                form_fields = ["id"]
                confirm_text = "Are you sure you want to delete this item?"

    Note:
        - Requires both form_fields and confirm_text to be specified
        - Will include CSRF token automatically
        - The form action URL is determined by the :py:attr:`attribute` method
          or attribute, which must be present on the model class.  If you need
          more sophisticated URL handling, subclass and override the
          :py:meth:`get_url` method.


    Keyword Args:
        form_fields: List of row attribute names to include as hidden fields
        confirm_text: Confirmation message to show before submission

    Raises:
        ImproperlyConfigured: If form_fields or confirm_text are not set

    """

    block: str = "action-form"
    tag: str = "form"

    form_fields: list[str] = []  # noqa: RUF012
    confirm_text: str | None = None

    def __init__(
        self,
        form_fields: list[str] | None = None,
        confirm_text: str | None = None,
        **kwargs,
    ):
        self.form_fields = form_fields or deepcopy(self.form_fields)
        self.confirm_text = confirm_text or self.confirm_text
        if not self.confirm_text:
            msg = "RowFormButton requires a 'confirm_text' argument or class attribute"
            raise ImproperlyConfigured(msg)
        if not self.form_fields:
            msg = "RowFormButton requires a 'form_fields' argument or class attribute"
            raise ImproperlyConfigured(msg)
        super().__init__(**kwargs)
        # Our superclass added self.text as a block; but that will not work for us,
        # since we want self.text to be the button text, so remove it.
        self.blocks = []
        self._attributes["method"] = "post"

    def get_confirm_text(self, row: Any) -> str:  # noqa: ARG002
        """
        Get the confirmation text for this button.

        Args:
            row: The row data this button is associated with

        Returns:
            str: The confirmation text to display

        """
        return self.confirm_text  # type: ignore[return-value]

    def bind(self, row: Any, table: Any, size: str | None = None) -> RowFormButton:
        """
        Bind this button to a specific row and create a form.

        Creates a complete form with:

        - CSRF token
        - Hidden fields with row data
        - Submit button with confirmation dialog

        Args:
            row: The row data to bind this button to
            table: The table instance that contains this row
            size: Optional size override for the button

        Returns:
            RowFormButton: A configured form button for the specified row

        """
        if not size:
            size = self.size
        action = deepcopy(self)
        action.add_block(
            HiddenInputBlock(input_name="csrfmiddlewaretoken", value=table.csrf_token)
        )
        action._attributes["action"] = self.get_url(row)
        for field in self.form_fields:
            action.add_block(
                HiddenInputBlock(input_name=field, value=getattr(row, field))
            )
        action.add_block(
            InputButton(
                text=self.text,
                color=self.color,
                size=size,
                confirm_text=self.get_confirm_text(row),
            )
        )
        return action


class RowEditButton(RowModelUrlButton):
    """
    A pre-configured button for editing a row's model instance.

    This is a convenience subclass of RowModelUrlButton configured specifically
    for edit operations. It looks for a get_update_url method on the model.

    Either subclass this, setting class attributes to create specific
    model-based buttons, or use it directly with keyword arguments to create
    buttons that link to model-specific URLs.  Note that all keyword
    arguments/class attributes from :py:class:`wildewidgets.Block`  and
    :py:class:`RowModelUrlButton` are also accepted.

    Attributes:
        attribute: The model method to call for URL ("get_update_url").  Your
            model must implement this method or attribute to return the URL for
            editing.
        text: Default button text ("Edit")
        color: Default Bootstrap color ("primary")

    Example:
        Direct usage with keyword arguments:

        .. code-block:: python

            from wildewidgets import RowEditButton

            # Simple usage - requires model.get_update_url() to exist
            button = RowEditButton()

            # Custom text
            button = RowEditButton(text="Modify")

        Subclassing for specific actions:

            class RowCustomEditButton(RowEditButton):
                attribute = "get_custom_update_url"
                text = "Modify"
                color = "warning"
                url_args = ["id"]

    """

    attribute: str = "get_update_url"
    text: str = "Edit"
    color: str = "primary"


class RowDeleteButton(RowFormButton):
    """
    A pre-configured button for deleting a row's model instance.

    This is a convenience subclass of RowFormButton configured specifically
    for delete operations. It submits the row's ID via POST to the URL
    returned by the model's ``get_delete_url`` method by default.

    Either subclass this, setting class attributes to create specific
    model-based buttons, or use it directly with keyword arguments to create
    buttons that link to model-specific URLs.  Note that all keyword
    arguments/class attributes from :py:class:`wildewidgets.Block`  and
    :py:class:`RowModelUrlButton` are also accepted.


    Important:
        Your model must implement a method or attribute named
        :py:attr:`attribute`, which must return the URL to submit the
        form to for deletion.  By default, this is
        :py:attr:`get_delete_url`.

        The form submitted will have by default the following fields:

        - ``csrf_token``: CSRF token for security
        - ``id``: The ID of the row's model instance

        Thus, your delete view should expect a POST request with these fields.

    Note:
        The confirmation text is generated using the row's string
        representation, so ensure your model has a meaningful `__str__` method.
        If you need a custom confirmation message, subclass and override
        :py:meth:`get_confirm_text`.

    Attributes:
        attribute: The model method or attribute to call/access for the URL
            ("get_delete_url")
        text: Default button text ("Delete")
        color: Default Bootstrap color ("outline-secondary")
        form_fields: Default field to submit (["id"])

    Example:
        Direct usage with keyword arguments:

        .. code-block:: python

            from wildewidgets import RowDeleteButton

            # Simple usage - requires model.get_delete_url() to exist
            button = RowDeleteButton()

        Subclassing for specific actions:

            class RowCustomDeleteButton(RowDeleteButton):
                attribute = "get_custom_delete_url"
                text = "Remove"
                color = "danger"
                form_fields = ["id", "name"]  # Submit additional fields

    """

    attribute: str = "get_delete_url"
    text: str = "Delete"
    color: str = "outline-secondary"
    form_fields: list[str] = ["id"]  # noqa: RUF012

    def get_confirm_text(self, row: Any) -> str:
        """
        Generate a confirmation message for deleting this row.

        Creates a message like 'Delete "Item Name"?' using the row's string
        representation.

        Args:
            row: The row data to create a confirmation message for

        Returns:
            str: The confirmation message to display

        """
        return f'Delete "{row!s}"?'


# -------------------------------
# Mixins
# -------------------------------


class ActionsButtonsBySpecMixin:
    """
    A mixin for :py:class:`wildewidgets.DataTable` classes and subclasses that
    adds action buttons based on specifications.

    This mixin allows you to add action buttons to the rightmost column of a
    table by specifying them as tuples that define their behavior. It provides a
    simple way to create standard actions without needing to define button
    classes.

    Actions are specified as tuples with these elements, in this order:

        1. Label (str): The button text
        2. URL name (str): Django URL pattern name, e.g., "app:item-view".  We'll
           reverse this to get the URL.
        3. Method (str, optional): "get" or "post" (default: "get")
        4. Color (str, optional): Bootstrap color class (default: "secondary")
        5. Field name (str, optional): ID field name (default: "id")
        6. JS function (str, optional): JavaScript function to call

    Example:
        .. code-block:: python

            from wildewidgets import ActionsButtonsBySpecMixin

            class MyDataTable(ActionsButtonsBySpecMixin, DataTable):
                actions = [
                    ("View", "app:item-view"),
                    ("Edit", "app:item-edit", "get", "primary"),
                    ("Delete", "app:item-delete", "post", "danger", "id",
                    "confirmDelete")
                ]

        This will add an "Actions" column with buttons for each specified action.
        Each button will be rendered with the appropriate URL and method based on the
        row data.

    Args:
        *args: Positional arguments for the parent class

    Keyword Args:
        actions: List of action specifications to add to the table
        action_button_size: Size of the action buttons (default: "normal")
        default_action_button_label: Default label for the action button
            (default: "View")
        default_action_button_color_class: Default color class for the action button
            (default: "secondary")
        **kwargs: Keyword arguments for the parent class

    """

    #: Per row action buttons.  If not ``False``, this will simply add a
    #: rightmost column  named ``Actions`` with a button named
    #: :py:attr:`default_action_button_label` which when clicked will take the
    #: user to the
    actions: Any = False
    #: How big should each action button be? One of ``normal``, ``btn-lg``, or
    #: ``btn-sm``.
    action_button_size: str = "normal"
    #: The label to use for the default action button
    default_action_button_label: str = "View"
    #: The Bootstrap color class to use for the default action buttons
    default_action_button_color_class: str = "secondary"

    def __init__(
        self,
        *args,
        actions: Any = None,
        action_button_size: str | None = None,
        default_action_button_label: str | None = None,
        default_action_button_color_class: str | None = None,
        **kwargs,
    ):
        self.actions = actions if actions is not None else self.actions
        self.action_button_size = action_button_size or self.action_button_size
        self.default_action_button_label = (
            default_action_button_label or self.default_action_button_label
        )
        self.default_action_button_color_class = (
            default_action_button_color_class or self.default_action_button_color_class
        )
        if self.action_button_size != "normal":
            self.action_button_size_class = f"btn-{self.action_button_size}"
        else:
            self.action_button_size_class = ""
        super().__init__(*args, **kwargs)

    def get_template_context_data(self, **kwargs) -> dict[str, Any]:
        """
        Add action column information to the template context.

        Args:
            **kwargs: Keyword arguments to update

        Returns:
            dict: Updated context dictionary with action column information

        """
        if self.actions:
            self.add_column(field="actions", searchable=False, sortable=False)
            kwargs["has_actions"] = True
            kwargs["action_column"] = len(self.column_fields) - 1
        else:
            kwargs["has_actions"] = False
        return super().get_template_context_data(**kwargs)  # type: ignore[misc]

    def get_content(self, **kwargs) -> str:
        """
        Get the rendered content for the table, including action columns.

        Args:
            **kwargs: Keyword arguments for content generation

        Returns:
            str: The HTML content for the table

        """
        if self.actions:
            self.add_column(field="actions", searchable=False, sortable=False)
        return super().get_content(**kwargs)  # type: ignore[misc]

    def get_action_button(
        self,
        row: Any,
        label: str,
        url_name: str,
        method: str = "get",
        color_class: str = "secondary",
        attr: str = "id",
        js_function_name: str | None = None,
    ) -> str:
        """
        Create an action button for a specific row using a Django URL name.

        Args:
            row: The row data this button is for
            label: Text to display on the button
            url_name: Django URL pattern name to link to
            method: HTTP method to use ("get" or "post")
            color_class: Bootstrap color class
            attr: Row attribute to use as ID parameter
            js_function_name: Optional JavaScript function to call on click

        Returns:
            str: HTML for the rendered button

        """
        if url_name:
            base = reverse(url_name)
            # TODO: This assumes we're using QueryStringKwargsMixin, which people
            # outside our group don't use
            url = f"{base}?{attr}={row.id}" if method == "get" else base
        else:
            url = "javascript:void(0)"
        return self.get_action_button_with_url(
            row, label, url, method, color_class, attr, js_function_name
        )

    def get_action_button_url_extra_attributes(self, row: Any) -> str:  # noqa: ARG002
        """
        Get additional URL query parameters for action buttons.

        Override this in subclasses to add custom query parameters.

        Args:
            row: The row data this button is for

        Returns:
            str: Additional URL query parameters

        """
        return ""

    def get_action_button_with_url(
        self,
        row: Any,
        label: str,
        url: str,
        method: str = "get",
        color_class: str = "secondary",
        attr: str = "id",
        js_function_name: str | None = None,
    ) -> str:
        """
        Create an action button for a row with a specific URL.

        Args:
            row: The row data this button is for
            label: Text to display on the button
            url: Complete URL to link to
            method: HTTP method to use ("get" or "post")
            color_class: Bootstrap color class
            attr: Row attribute to use as ID parameter for POST forms
            js_function_name: Optional JavaScript function to call on click

        Returns:
            str: HTML for the rendered button

        """
        url_extra = self.get_action_button_url_extra_attributes(row)
        if url_extra:
            url = f"{url}&{url_extra}"
        if method == "get":
            if js_function_name:
                link_extra = f'onclick="{js_function_name}({row.id});"'
            else:
                link_extra = ""
            return f'<a href="{url}" class="btn btn-{color_class} {self.action_button_size_class} me-2" {link_extra}>{label}</a>'  # noqa: E501
        token_input = f'<input type="hidden" name="csrfmiddlewaretoken" value="{self.csrf_token}">'  # noqa: E501
        id_input = f'<input type="hidden" name="{attr}" value="{row.id}">'
        button = f'<input type=submit value="{label}" class="btn btn-{color_class} {self.action_button_size_class} me-2">'  # noqa: E501
        return f'<form class="form form-inline" action={url} method="post">{token_input}{id_input}{button}</form>'  # noqa: E501

    def get_conditional_action_buttons(self, row: Any) -> str:  # noqa: ARG002
        """
        Get additional action buttons based on row data.

        Override this in subclasses to add buttons conditionally based on
        row attributes.

        Args:
            row: The row data to evaluate

        Returns:
            str: HTML for additional buttons

        """
        return ""

    def render_actions_column(self, row: Any, column: str) -> str:  # noqa: ARG002
        """
        Render all action buttons for a specific row.

        This method generates the complete HTML for the actions column, including:
        1. The default view button if the model has get_absolute_url
        2. All buttons defined in the actions specification
        3. Any conditional buttons from get_conditional_action_buttons

        Args:
            row: The row data to render buttons for
            column: The column name (ignored, always "actions")

        Returns:
            str: Complete HTML for the actions column

        """
        response = '<div class="d-flex flex-row justify-content-end">'
        if hasattr(row, "get_absolute_url"):
            if callable(row.get_absolute_url):
                url = row.get_absolute_url()
            else:
                url = row.get_absolute_url
            view_button = self.get_action_button_with_url(
                row,
                self.default_action_button_label,
                url,
                color_class=self.default_action_button_color_class,
            )
            response += view_button
        if not isinstance(self.actions, bool):
            for action in self.actions:
                if not len(action) > 1:
                    continue
                label = action[0]
                url_name = action[1]
                method = action[2] if len(action) > 2 else "get"  # noqa: PLR2004
                color_class = action[3] if len(action) > 3 else "secondary"  # noqa: PLR2004
                attr = action[4] if len(action) > 4 else "id"  # noqa: PLR2004
                js_function_name = action[5] if len(action) > 5 else ""  # noqa: PLR2004
                response += self.get_action_button(
                    row,
                    label,
                    url_name,
                    method,
                    color_class,
                    attr,
                    js_function_name,
                )
        response += self.get_conditional_action_buttons(row)
        response += "</div>"
        return response


class ActionButtonBlockMixin:
    """
    A mixin for :py:class:`wildwidgets.DataTable` classes that adds action
    buttons using :py:class:`RowActionButton` classes.

    This mixin provides a more object-oriented approach to adding action buttons
    compared to :py:class:`ActionsButtonsBySpecMixin`. It uses
    :py:class:`RowActionButton` instances for greater flexibility and
    extensibility.

    Important:
        Typically, you will not use this directly, but rather use
        :py:class:`wildewidgets.StandardActionButtonModelTable` or
        :py:class:`wildewidgets.LookupModelTable`, which already
        has this built in as a mixin.

    Example:
        .. code-block:: python

            from wildewidgets import ActionButtonBlockMixin, DataTable

            class MyTable(ActionButtonBlockMixin, DataTable):
                actions = [
                    RowEditButton(),
                    RowDeleteButton()
                ]
                ...

    Args:
        *args: Positional arguments for the parent class

    Keyword Args:
        actions: List of :py:class:`RowActionButton` instances to add to the
            "Actions" column. If not provided, a default "View" button is used.
        button_size: Size of the action buttons (default: None)
        justify: Justification of the action buttons in the "Actions" column
            (default: "end")

    """

    #: If not ``None``, make all per-row action buttons be this size.
    button_size: str | None = None
    #: The justification of the action buttons in the "Actions" column.
    justify: Literal["start", "center", "end"] = "end"

    #: A list of :py:class:`RowActionButton` subclasses to display in the
    #: "Actions" column.
    actions: list[RowActionButton] = [RowModelUrlButton(text="View", color="secondary")]  # noqa: RUF012

    def __init__(
        self,
        *args,
        actions: list[RowActionButton] | None = None,
        button_size: str | None = None,
        justify: Literal["start", "center", "end"] | None = None,
        **kwargs,
    ):
        self.actions = actions if actions is not None else deepcopy(self.actions)
        self.button_size = button_size or self.button_size
        self.justify = justify or self.justify
        super().__init__(*args, **kwargs)

    def get_actions(self) -> list[RowActionButton]:
        """
        Get the list of action buttons to display.

        Override this in subclasses to dynamically determine which
        buttons to show.

        Returns:
            list[RowActionButton]: List of action button instances

        """
        return self.actions

    def get_content(self, **kwargs) -> str:
        """
        Get the rendered content for the table, including action columns.

        Args:
            **kwargs: Keyword arguments for content generation

        Returns:
            str: The HTML content for the table

        """
        if self.get_actions():
            self.add_column(field="actions", searchable=False, sortable=False)
        return super().get_content(**kwargs)  # type: ignore[misc]

    def render_actions_column(self, row: Any, column: str) -> str:  # noqa: ARG002
        """
        Render all action buttons for a specific row.

        Creates a horizontal layout containing all visible action buttons
        for the given row.

        Args:
            row: The row data to render buttons for
            column: The column name (ignored, always "actions")

        Returns:
            str: Complete HTML for the actions column

        Note:
            Buttons are only shown if their is_visible method returns True
            for the current user.

        """
        container = HorizontalLayoutBlock(justify=self.justify)
        for action in self.get_actions():
            button = action.bind(row, self, size=self.button_size)
            user = None
            if hasattr(self, "request") and self.request is not None:
                user = self.request.user
            if not button.is_visible(row, user):
                continue
            button.add_class("me-2")
            container.add_block(button)
        button.remove_class("me-2")
        return str(container)


class StandardModelActionButtonBlockMixin(ActionButtonBlockMixin):
    """
    A ready-to-use mixin providing standard "Edit" and "Delete" buttons for
    model tables.

    This mixin provides a convenient configuration of
    :py:class:`ActionButtonBlockMixin` with "Edit" and "Delete" buttons
    pre-configured. It's a quick way to add these common actions to any
    DataTable.

    Note:
        This is used by :py:class:`wildewidgets.LookupModelTable`.

    Requirements:
        - Your model must implement ``get_absolute_url``, ``get_update_url`` and
          ``get_delete_url`` methods

    Example:
        .. code-block:: python

            from wildewidgets import StandardModelActionButtonBlockMixin, DataTable
            from myapp.models import MyModel

            class MyTable(StandardModelActionButtonBlockMixin, DataTable):
                model = MyModel
                fields = ["name", "description"]
                # No additional configuration needed for basic edit/delete buttons

    """

    actions: list[RowActionButton] = [  # noqa: RUF012
        RowModelUrlButton(text="Edit", color="primary", attribute="get_update_url"),
        # TODO: change this to RowFormButton
        RowModelUrlButton(
            text="delete",
            color="outline-secondary",
            attribute="get_delete_url",
        ),
    ]



from __future__ import annotations

import random
from typing import Any, ClassVar

from django import template
from django.urls import reverse

from wildewidgets.views import WidgetInitKwargsMixin


class BasicMenu(WidgetInitKwargsMixin):
    """
    Basic menu widget.

    A basic menu requires only one class attribute defined, :py:attr:`items`.

    Example:
        .. code-block:: python

            from wildewidgets import BasicMenu

            class MainMenu(BasicMenu):
                items = [
                    ('Users', 'core:home'),
                    ('Uploads','core:uploads'),
                ]

    """

    #: The Django template file to use for rendering the menu.
    template_file: str = "wildewidgets/menu.html"
    #: The CSS classes to use for the navbar.  This can be set to
    navbar_classes: str = "navbar-expand-lg navbar-light"
    #: The CSS class to use for the container that holds the navbar.
    container: str = "container-lg"
    #: The image to use for the brand logo.  If this is set, the brand text will
    #: not be displayed.  If you want to use a brand text, set this to
    #: ``None``.  The image should be a URL or a static file path.
    brand_image: str | None = None
    #: The CSS width to use for the brand image.  This is only used if
    #: :py:attr:`brand_image` is set.  If you want the image to be responsive,
    #: set this to ``100%``.
    brand_image_width: str = "100%"
    #: The text to use for the brand.
    brand_text: str | None = None
    #: The URL to use for the brand link.  If this is set, the brand image or
    #: text will be used as the link.
    brand_url: str = "#"
    #: A list of tuples that define the items to list in the menu, where the
    #: first element is the menu item text and the second element is the URL
    #: name.  A third optional element in the tuple can be a dictionary of get
    #: arguments
    items: ClassVar[
        list[tuple[str, str | list[tuple[str, str | None]] | dict[str, str] | None]]
    ] = []

    def __init__(self, *args: Any, **kwargs: Any) -> None:  # noqa: ARG002
        self.menu: dict[str, Any] = {}
        self.active: str | None = None
        if args:
            self.active_hierarchy: list[str] = args[0].split("/")
        else:
            self.active_hierarchy = []

    def build_menu(self) -> None:
        """
        Build the menu structure based on the active hierarchy.

        This method processes the items defined in the class attribute and creates
        a structured representation of the menu. It handles both regular menu items
        and submenus, and marks the appropriate items as active based on the
        active_hierarchy.

        The method:

        1. Processes each item in the items list
        2. Creates appropriate data structures for regular items and submenus
        3. Marks items as active if they match the active_hierarchy
        4. Adds query parameters if specified in the item definition

        Note:
            This method is called automatically by get_content() and doesn't need
            to be called directly.

        """
        if len(self.active_hierarchy) > 0:
            for item in self.items:
                data: dict[str, Any] = {}
                if isinstance(item[1], str):
                    data["url"] = reverse(item[1])
                    data["extra"] = ""
                    data["kind"] = "item"

                    if len(item) > 2 and item[2] is not None:  # noqa: PLR2004
                        extra = item[2]
                        if isinstance(extra, dict):
                            extra_list = []
                            for k, v in extra.items():
                                extra_list.append(f"{k}={v}")
                            extra_str = f"?{'&'.join(extra_list)}"
                            data["extra"] = extra_str
                elif isinstance(item[1], list):
                    if len(self.active_hierarchy) > 1:
                        submenu_active = self.active_hierarchy[1]
                    else:
                        submenu_active = None
                    data = self.parse_submemu(item[1], submenu_active)

                self.add_menu_item(item[0], data, item[0] == self.active_hierarchy[0])

    def add_menu_item(
        self, title: str, data: dict[str, Any], active: bool = False
    ) -> None:
        """
        Add a menu item to the menu structure.

        This method adds a new item to the menu with the specified title and data,
        and optionally marks it as the active item.

        Args:
            title: The text to display for the menu item
            data: Dictionary containing the item's configuration (url, kind, etc.)

        Keyword Args:
            active: Whether this item should be marked as active

        """
        self.menu[title] = data
        if active:
            self.active = title

    def parse_submemu(
        self, items: list[tuple[str, str | None]], submenu_active: str | None
    ) -> dict[str, Any]:
        """
        Parse and create a submenu structure.

        This method processes a list of submenu items and creates a structured
        representation of the submenu. It handles regular items and dividers,
        and marks the appropriate item as active based on the submenu_active parameter.

        Args:
            items: List of tuples defining submenu items (title, url, [extra])

        Keyword Args:
            submenu_active: The name of the submenu item to mark as active

        Returns:
            dict: A dictionary containing the structured submenu data

        """
        data: dict[str, Any] = {"kind": "submenu"}
        sub_menu_items: list[dict[str, Any]] = []
        for item in items:
            if not isinstance(item, tuple):
                continue
            if item[0] == "divider":
                subdata: dict[str, Any] = {"divider": True}
            else:
                subdata = {
                    "title": item[0],
                    "url": reverse(item[1]) if item[1] else "#",
                    "extra": "",
                    "divider": False,
                    "active": item[0] == submenu_active,
                }

            if len(item) > 2:  # noqa: PLR2004
                subdata["extra"] = self.convert_extra(item[2])
            sub_menu_items.append(subdata)

        data["items"] = sub_menu_items
        return data

    def get_content(self, **kwargs: Any) -> str:  # noqa: ARG002
        """
        Render the menu as HTML content.

        This method:

        1. Builds the menu structure
        2. Creates a context dictionary with all necessary data
        3. Renders the template with the context

        Keyword Args:
            **kwargs: Additional arguments (unused)

        Returns:
            str: The rendered HTML content for the menu

        """
        self.build_menu()
        context = {
            "menu": self.menu,
            "active": self.active,
            "navbar_classes": self.navbar_classes,
            "navbar_container": self.container,
            "brand_image": self.brand_image,
            "brand_image_width": self.brand_image_width,
            "brand_text": self.brand_text,
            "brand_url": self.brand_url,
            "vertical": "navbar-vertical" in self.navbar_classes,
            "target": random.randrange(0, 10000),  # noqa: S311
        }
        html_template = template.loader.get_template(self.template_file)
        return html_template.render(context)

    def __str__(self) -> str:
        """
        Return the string representation of the menu.

        This method allows the menu to be used directly in templates via the
        {{ menu }} syntax.

        Returns:
            str: The rendered HTML content for the menu

        """
        return self.get_content()


class DarkMenu(BasicMenu):
    """
    A dark-themed navigation menu.

    This class extends the BasicMenu with dark styling using Bootstrap's dark
    navbar classes.  It provides a dark background with light text, suitable for
    applications with darker themes.

    Attributes:
        navbar_classes: Bootstrap classes for styling the navbar with dark theme

    Example:
        .. code-block:: python

            from wildewidgets import DarkMenu

            class MainDarkMenu(DarkMenu):
                items = [
                    ('Dashboard', 'dashboard'),
                    ('Reports', 'reports'),
                ]

    """

    navbar_classes = "navbar-expand-lg navbar-dark bg-secondary"


class VerticalDarkMenu(BasicMenu):
    """
    A vertical dark-themed navigation menu.

    This class extends the BasicMenu with dark styling and vertical orientation.
    It's particularly useful for sidebar navigation in applications with darker
    themes.

    Attributes:
        navbar_classes: Bootstrap classes for styling a vertical dark navbar

    Example:
        .. code-block:: python

            from wildewidgets import VerticalDarkMenu

            class SidebarMenu(VerticalDarkMenu):
                items = [
                    ('Profile', 'user_profile'),
                    ('Settings', 'user_settings'),
                    ('Logout', 'logout'),
                ]

    """

    navbar_classes = "navbar-vertical navbar-expand-lg navbar-dark"


class LightMenu(BasicMenu):
    """
    A light-themed navigation menu.

    This class extends the BasicMenu with light styling, which is also the
    default for BasicMenu.  It provides a light background with dark text,
    suitable for standard application layouts.

    Attributes:
        navbar_classes: Bootstrap classes for styling the navbar with light theme

    Example:
        .. code-block:: python

            from wildewidgets import LightMenu

            class MainLightMenu(LightMenu):
                items = [
                    ('Home', 'home'),
                    ('About', 'about'),
                    ('Contact', 'contact'),
                ]

    """

    navbar_classes = "navbar-expand-lg navbar-light"


class MenuMixin:
    """
    A mixin for adding menu support to Django class-based views.

    This mixin provides methods for including primary and secondary navigation menus
    in Django views. It automatically adds menu instances to the template context,
    making it easy to integrate navigation in templates.

    To use this mixin:

    1. Define menu classes by subclassing :py:class:`BasicMenu` or its variants
    2. Add this mixin to your view classes
    3. Set class attributes to specify which menus to use and which items to activate

    You must define the :py:attr:`menu_class` and the :py:attr:`menu_item`, but
    :py:attr:`submenu_class` and the :py:attr:`submenu_item` are optional. If
    you do not set them, no secondary menu will be displayed.

    Example:
        With no secondary menu:

        .. code-block:: python

            from django.views.generic import TemplateView
            from wildewidgets import BasicMenu, MenuMixin

            class MainMenu(BasicMenu):
                items = [
                    ('Home', 'home'),
                    ('About', 'about'),
                    ('Contact', 'contact'),
                ]

            class DashboardView(MenuMixin, TemplateView):
                template_name = "dashboard.html"
                menu_class = MainMenu
                menu_item = "Home"

        With a secondary menu:

        .. code-block:: python

            from django.views.generic import TemplateView
            from wildewidgets import BasicMenu, LightMenu, MenuMixin

            class MainMenu(BasicMenu):
                items = [
                    ('Home', 'home'),
                    ('About', 'about'),
                    ('Contact', 'contact'),
                ]

            class DashboardSubmenu(LightMenu):
                items = [
                    ('Overview', 'dashboard:overview'),
                    ('Statistics', 'dashboard:statistics'),
                    ('Settings', 'dashboard:settings'),
                ]

            class DashboardView(MenuMixin, TemplateView):
                template_name = "dashboard.html"
                menu_class = MainMenu
                menu_item = "Home"
                submenu_class = DashboardSubmenu
                submenu_item = "Overview"

    """

    #: The primary menu class to use for the view.
    #: This should be a subclass of BasicMenu or one of its variants.
    #: If set to None, no primary menu will be displayed.
    menu_class: type[BasicMenu] | None = None
    #: The active item for the primary menu.
    menu_item: str | None = None
    #: The secondary menu class to use for the view.
    submenu_class: type[BasicMenu] | None = None
    #: The active item for the secondary menu.
    submenu_item: str | None = None

    def get_menu_class(self) -> type[BasicMenu] | None:
        """
        Get the class to use for the primary menu.

        This method returns the menu_class attribute by default, but can be
        overridden to provide dynamic menu class selection.

        Returns:
            The menu class to use, or None if no menu should be displayed

        """
        return self.menu_class

    def get_menu_item(self) -> str | None:
        """
        Get the active item for the primary menu.

        This method returns the menu_item attribute by default, but can be
        overridden to provide dynamic active item selection.

        Returns:
            The name of the menu item to mark as active, or None if no item
            should be active

        """
        return self.menu_item

    def get_menu(self) -> BasicMenu | None:
        """
        Get an instance of the primary menu.

        This method instantiates the menu class with the active item.

        Returns:
            An instance of the menu, or None if no menu class is specified

        """
        menu_class = self.get_menu_class()
        if menu_class:
            menu_item = self.get_menu_item()
            return menu_class(menu_item)  # pylint: disable=not-callable
        return None

    def get_submenu_class(self) -> type[BasicMenu] | None:
        """
        Get the class to use for the secondary menu.

        This method returns the submenu_class attribute by default, but can be
        overridden to provide dynamic submenu class selection.

        Returns:
            The submenu class to use, or None if no submenu should be displayed

        """
        return self.submenu_class

    def get_submenu_item(self) -> str | None:
        """
        Get the active item for the secondary menu.

        This method returns the submenu_item attribute by default, but can be
        overridden to provide dynamic active item selection.

        Returns:
            The name of the submenu item to mark as active, or None if no item
            should be active

        """
        return self.submenu_item

    def get_submenu(self) -> BasicMenu | None:
        """
        Get an instance of the secondary menu.

        This method instantiates the submenu class with the active item.

        Returns:
            An instance of the submenu, or None if no submenu class is specified

        """
        submenu_class = self.get_submenu_class()
        if submenu_class:
            submenu_item = self.get_submenu_item()
            return submenu_class(submenu_item)  # pylint: disable=not-callable
        return None

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        """
        Add menu instances to the template context.

        This method adds the primary and secondary menu instances to the context
        if they are available.

        Args:
            **kwargs: Additional context variables

        Returns:
            The updated template context with menu instances

        """
        menu = self.get_menu()
        submenu = self.get_submenu()
        if menu:
            kwargs["menu"] = menu
        if submenu:
            kwargs["submenu"] = submenu
        return super().get_context_data(**kwargs)  # type: ignore[misc]



from __future__ import annotations

import importlib
from pathlib import Path
from typing import Any

from django.apps import apps
from django.http import Http404, HttpRequest, HttpResponseBase, JsonResponse
from django.views.generic import View
from django.views.generic.base import TemplateView

from .mixins import JSONResponseMixin, WidgetInitKwargsMixin


class JSONResponseView(JSONResponseMixin, TemplateView):  # type: ignore[misc]
    """
    A view that renders a template and returns the result as a JSON response.

    This view combines Django's TemplateView with JSONResponseMixin to provide a
    simple way to return template-rendered content within a JSON response. This
    is useful for AJAX requests that need to return HTML fragments within a JSON
    structure.

    When subclassing, you typically need to override the template_name attribute
    and optionally the get_context_data method.

    Attributes:
        template_name: The template to render (must be set by subclasses)

    Example:
        .. code-block:: python

            class MyJSONView(JSONResponseView):
                template_name = "my_template.html"

                def get_context_data(self, **kwargs):
                    context = super().get_context_data(**kwargs)
                    context['extra_data'] = get_some_data()
                    return context

    """


class JSONDataView(View):
    """
    A view that returns JSON data in response to HTTP requests.

    This class provides a simple framework for views that need to return JSON data.
    It handles the HTTP request/response cycle and serializes the context data
    to JSON. By default, it only responds to GET requests.

    To use this class, subclass it and override the get_context_data method to
    provide the data you want to return as JSON.

    Example:
        .. code-block:: python

            class UserDataView(JSONDataView):
                def get_context_data(self, **kwargs):
                    user_id = self.kwargs.get('user_id')
                    user = User.objects.get(id=user_id)
                    return {
                        'username': user.username,
                        'email': user.email,
                        'date_joined': user.date_joined.isoformat()
                    }

    """

    def get(self, request: HttpRequest, *args, **kwargs) -> JsonResponse:  # noqa: ARG002
        """
        Handle GET requests by returning JSON data.

        This method retrieves the context data and returns it as a JSON response.

        Args:
            request: The HTTP request object
            *args: Additional positional arguments
            **kwargs: Additional keyword arguments

        Returns:
            JsonResponse: HTTP response containing the JSON-serialized context data

        """
        context = self.get_context_data()
        return self.render_to_response(context)

    def get_context_data(self, **kwargs) -> dict[str, Any]:  # noqa: ARG002
        """
        Get the data to include in the JSON response.

        Override this method to provide the data you want to include in the response.
        By default, it returns an empty dictionary.

        Args:
            **kwargs: Additional context data

        Returns:
            dict: The data to serialize to JSON

        """
        return {}

    def render_to_response(self, context, **response_kwargs) -> JsonResponse:  # noqa: ARG002
        """
        Create a JSON response from the context data.

        This method serializes the context data to JSON and returns it as an
        HTTP response.

        Args:
            context: The data to serialize to JSON
            **response_kwargs: Additional response parameters (unused)

        Returns:
            JsonResponse: HTTP response containing the JSON-serialized context data

        """
        return JsonResponse(context)


class WildewidgetDispatch(WidgetInitKwargsMixin, View):
    """
    A view that dynamically dispatches requests to widget classes.

    This view acts as a central dispatcher for widget AJAX requests. It examines
    the request parameters to determine which widget class to instantiate, then
    delegates the request handling to that widget's dispatch method.

    The view searches for widget classes in all installed Django apps by looking
    for a 'wildewidgets.py' file or a 'wildewidgets' directory within each app.

    This approach allows widgets to handle their own AJAX requests without
    requiring explicit URL routing for each widget type.

    Example usage in URLs:
        .. code-block:: python

            path(
                'wildewidget/',
                WildewidgetDispatch.as_view(),
                name='wildewidget_dispatch'
            )

    Example client-side code:
        .. code-block:: javascript

            $.ajax({
                url: '/wildewidget/',
                data: {
                    'wildewidgetclass': 'MyWidget',
                    'extra_data': encodeURIComponent(JSON.stringify({
                        args: [],
                        kwargs: {param1: 'value1'}
                    }))
                }
            });
    """

    def dispatch(  # type: ignore[override]
        self, request: HttpRequest, *args, **kwargs
    ) -> HttpResponseBase | Http404:
        """
        Dispatch the request to the appropriate widget class.

        This method:

        1. Extracts the widget class name from the request
        2. Searches for the widget class in all installed apps
        3. Instantiates the widget class with the provided arguments
        4. Delegates to the widget's dispatch method

        Args:
            request: The HTTP request object
            *args: Additional positional arguments

        Keyword Arguments:
            **kwargs: Additional keyword arguments

        Returns:
            HttpResponseBase: The response from the widget's dispatch method
            Http404: If the widget class cannot be found

        Note:
            The request must include:
            - 'wildewidgetclass': The name of the widget class to instantiate

            The request may include:
            - 'csrf_token': Optional CSRF token for protected requests
            - 'extra_data': Optional base64 encoded JSON encoded string
              containing a dict with  'args' and 'kwargs' keys for widget
              initialization

        """
        wildewidgetclass = request.GET.get("wildewidgetclass", None)
        csrf_token = request.GET.get("csrf_token", "")
        if wildewidgetclass:
            configs = apps.get_app_configs()
            for config in configs:
                check_file = Path(config.path) / "wildewidgets.py"
                check_dir = Path(config.path) / "wildewidgets"
                if check_file.is_file() or check_dir.is_dir():
                    module = importlib.import_module(f"{config.name}.wildewidgets")
                    if hasattr(module, wildewidgetclass):
                        class_ = getattr(module, wildewidgetclass)
                        extra_data = self.get_decoded_extra_data(request)
                        initargs = extra_data.get("args", [])
                        initkwargs = extra_data.get("kwargs", {})
                        instance = class_(*initargs, **initkwargs)
                        instance.request = request
                        instance.csrf_token = csrf_token
                        instance.args = initargs
                        instance.kwargs = initkwargs
                        return instance.dispatch(request, *args, **kwargs)
        msg = "Not Found: Wildewidget class not found"
        raise Http404(msg)



import os

# These are the env vars that get retrieved with getenv(). They must be set to
# avoid raising UnsetEnvironmentVariable.
os.environ["DB_NAME"] = "fake"
os.environ["DB_USER"] = "fake"
os.environ["DB_PASSWORD"] = "fake"  # noqa: S105
os.environ["DB_HOST"] = "fake"
os.environ["DB_PORT"] = "fake"
os.environ["CACHE"] = "fake"
# These are set to True here ONLY so the static resouces for
# django-debug-toolbar will be collected during image build.
os.environ["DEVELOPMENT"] = "True"
os.environ["ENABLE_DEBUG_TOOLBAR"] = "True"

# noinspection PyUnresolvedReferences
from .settings import *  # noqa: F403



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from django.core.exceptions import ImproperlyConfigured

from .base import Block
from .forms import CrispyFormWidget

if TYPE_CHECKING:
    from django.forms import Form


class ModalWidget(Block):
    """
    Renders a Bootstrap 5 Modal dialog.

    This widget creates a Bootstrap modal dialog with a header, body, and footer.
    The modal can be triggered by a button or link with a data-target attribute
    matching the modal's ID.

    Example:
        .. code-block:: python

            from wildewidgets import ModalWidget, Block

            # Create a simple modal
            modal = ModalWidget(
                modal_id="example-modal",
                modal_title="Important Information",
                modal_body=Block("This is the modal content"),
                modal_size="lg"
            )

    Attributes:
        template_name: Path to the template for rendering the modal
        modal_id: Unique identifier for the modal (used in triggering elements)
        modal_title: Text to display in the modal header
        modal_body: Content to display in the modal body (typically a Block)
        modal_size: Size of the modal dialog (None, 'sm', 'lg', or 'xl')

    Keyword Args:
        modal_id: The CSS ID of the modal
        modal_title: The title of the modal
        modal_body: The body of the modal, any Block
        modal_size: The size of the modal (None, 'sm', 'lg', or 'xl')
        **kwargs: Additional attributes passed to the parent Block class

    """

    template_name: str = "wildewidgets/modal.html"
    modal_id: str | None = None
    modal_title: str | None = None
    modal_body: str | None = None
    modal_size: str | None = None

    def __init__(
        self,
        modal_id: str | None = None,
        modal_title: str | None = None,
        modal_body: str | None = None,
        modal_size: str | None = None,
        **kwargs: Any,
    ):
        self.modal_id = modal_id or self.modal_id
        self.modal_title = modal_title or self.modal_title
        self.modal_body = modal_body or self.modal_body
        self.modal_size = modal_size or self.modal_size
        super().__init__(**kwargs)

    def get_context_data(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """
        Prepare the context data for the modal template.

        This method adds the modal-specific attributes to the template context
        to be used in rendering the modal dialog.

        Args:
            *args: Variable length argument list
            **kwargs: Arbitrary keyword arguments

        Returns:
            dict: The updated context dictionary with modal attributes

        """
        kwargs = super().get_context_data(*args, **kwargs)
        kwargs["modal_id"] = self.modal_id
        kwargs["modal_title"] = self.modal_title
        kwargs["modal_body"] = self.modal_body
        kwargs["modal_size"] = self.modal_size
        return kwargs


class CrispyFormModalWidget(ModalWidget):
    """
    A modal dialog containing a Django form rendered with django-crispy-forms.

    This specialized modal automatically places a form in the modal body,
    making it easy to create form dialogs for user input. It handles the
    rendering of the form using the CrispyFormWidget.

    Example:
        .. code-block:: python

            from django import forms
            from wildewidgets import CrispyFormModalWidget

            class ContactForm(forms.Form):
                name = forms.CharField()
                email = forms.EmailField()
                message = forms.CharField(widget=forms.Textarea)

            # Create a modal with the form
            modal = CrispyFormModalWidget(
                modal_id="contact-modal",
                modal_title="Contact Us",
                form_class=ContactForm
            )

    Attributes:
        form_class: The form class to instantiate and render in the modal
        form: An already instantiated form to render in the modal

    Args:
        form: An instantiated form to render in the modal
        form_class: A form class to instantiate and render in the modal
        **kwargs: Additional arguments passed to the parent ModalWidget class
            (modal_id, modal_title, modal_size)

    Raises:
        ImproperlyConfigured: If neither form nor form_class is provided

    """

    form_class: type[Form] | None = None
    form: Form | None = None

    def __init__(
        self,
        form: Form | None = None,
        form_class: type[Form] | None = None,
        **kwargs: Any,
    ):
        if not form:
            form = self.form
        if not form_class:
            form_class = self.form_class

        if form_class:
            modal_form = form_class()
        elif form:
            modal_form = form
        else:
            msg = "Either 'form_class' or 'form' must be set"
            raise ImproperlyConfigured(msg)
        modal_body = CrispyFormWidget(form=modal_form)
        kwargs["modal_body"] = modal_body
        super().__init__(**kwargs)



from django.db import models

F = models.Field


class Measurement(models.Model):
    name: F = models.CharField(max_length=128)
    time: F = models.DateTimeField(
        "Time", help_text="The date and time the measurement was made."
    )
    pressure: F = models.DecimalField(max_digits=8, decimal_places=2)
    temperature: F = models.DecimalField(max_digits=8, decimal_places=2)
    restricted: F = models.BooleanField(default=False)
    open: F = models.BooleanField(default=True)

    def __str__(self) -> str:
        return f"Measurement({self.name}, {self.time})"



from django.contrib.auth import get_user_model
from django.core.management import call_command
from django.db import migrations

measurement_fixture = 'measurements'
book_manager_fixture = 'book_manager'
user_fixture = 'users'


def load_fixture(apps, schema_editor):
    call_command('loaddata', measurement_fixture, app_label='core')
    call_command('loaddata', user_fixture, app_label='core')
    call_command('loaddata', book_manager_fixture, app_label='core')


def unload_fixture(apps, schema_editor):
    Measurement = apps.get_model("core", "Measurement")
    Measurement.objects.all().delete()
    Publisher = apps.get_model("book_manager", "Publisher")
    Publisher.objects.all().delete()
    Binding = apps.get_model("book_manager", "Binding")
    Binding.objects.all().delete()
    Book = apps.get_model("book_manager", "Book")
    Book.objects.all().delete()
    Author = apps.get_model("book_manager", "Author")
    Author.objects.all().delete()
    Shelf = apps.get_model("book_manager", "Shelf")
    Shelf.objects.all().delete()
    Reading = apps.get_model("book_manager", "Reading")
    Reading.objects.all().delete()
    User = get_user_model()
    User.objects.all().delete()


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(load_fixture, reverse_code=unload_fixture),
    ]